# -*- coding: utf-8 -*-
"""
Unitary quaternions [#]_ are used when representing an attitude. They can be
updated via integration of angular rate measurements of a gyroscope.

The easiest way to do so is by integrating the differential
equation for a local rotation rate :cite:p:`sola2017quaternion`.

In a kinematic system, the angular velocity :math:`\\boldsymbol\\omega` of a
rigid body at any instantaneous time is described with respect to a fixed frame
coinciding instantaneously with its body frame. Thus, this angular
velocity is *in terms of* the body frame :cite:p:`jia2024`.

Accumulating rotation over time in quaternion form is done by integrating the
differential equation of :math:`\\mathbf{q}` with a defined rotation rate.
This constant augmentation is sometimes termed the **Attitude Propagation**.

.. math::
    \\hat{\\mathbf{q}}_t = \\mathbf{q}_{t-1} + \\int_{t-1}^t\\boldsymbol\\omega\\, dt

Exact closed-form solutions of the integration are not available. Thus, an
approximation method is required. Besides, numerical methods don't give a
continuous solution for :math:`\\mathbf{q}`, but a discrete set of values can:
:math:`\\mathbf{q}_t`, :math:`n=1,2,\\dots`

In the simplest practical case, the angular rates are measured by `gyroscopes
<https://en.wikipedia.org/wiki/Gyroscope>`_, reading instantaneous angular
velocities, :math:`\\boldsymbol\\omega(t_n)
=\\begin{bmatrix}\\omega_x&\\omega_y&\\omega_z\\end{bmatrix}^T`,
in *rad/s*, at discrete times :math:`t_n = n\\Delta t` in the local sensor
frame.

The parameter :math:`\\Delta t` is called *time step* or *step size* of the
numerical integration.

Quaternion Derivative
---------------------

We start by expressing our angular velocity in quaternion form, so that we can
use the quaternion derivative to integrate it and estimate the new attitude.

An orientation (attitude) is described with a quaternion :math:`\\mathbf{q} (t)`
at a time :math:`t`, and with :math:`\\mathbf{q} (t+\\Delta t)` at a time
:math:`t+\\Delta t`. This is after a rotation change :math:`\\Delta\\mathbf{q}`
during :math:`\\Delta t` seconds is performed on the local frame
:cite:p:`jia2024`.

This rotation change about the instantaneous axis
:math:`\\mathbf{u}=\\frac{\\boldsymbol\\omega}{\\|\\boldsymbol\\omega\\|}`
through the angle :math:`\\theta=\\|\\boldsymbol\\omega\\|\\Delta t` can be
described by a quaternion too:

.. math::
    \\begin{array}{rcl}
    \\Delta\\mathbf{q} &=& \\cos\\frac{\\theta}{2} + \\mathbf{u}\\sin\\frac{\\theta}{2} \\\\
    &=& \\cos\\frac{\\|\\boldsymbol\\omega\\|\\Delta t}{2} + \\frac{\\boldsymbol\\omega}{\\|\\boldsymbol\\omega\\|}\\sin\\frac{\\|\\boldsymbol\\omega\\|\\Delta t}{2}
    \\end{array}

implying that :math:`\\mathbf{q}(t+\\Delta t)=\\Delta\\mathbf{qq}(t)`. To
obtain the derivative we consider :math:`\\mathbf{q} = \\mathbf{q}(t)` as the
original state at a time :math:`t`, and the difference between states as:

.. math::
    \\begin{array}{rcl}
    \\mathbf{q}(t+\\Delta t)-\\mathbf{q}(t)
    &=& \\Big(\\cos\\frac{\\|\\boldsymbol\\omega\\|\\Delta t}{2} + \\frac{\\boldsymbol\\omega}{\\|\\boldsymbol\\omega\\|}\\sin\\frac{\\|\\boldsymbol\\omega\\|\\Delta t}{2}\\Big)\\mathbf{q} - \\mathbf{q} \\\\
    &=& \\Big(-2\\sin^2\\frac{\\|\\boldsymbol\\omega\\|\\Delta t}{4} + \\frac{\\boldsymbol\\omega}{\\|\\boldsymbol\\omega\\|}\\sin\\frac{\\|\\boldsymbol\\omega\\|\\Delta t}{2}\\Big)\\mathbf{q}
    \\end{array}

The development of the `time-derivative <https://en.wikipedia.org/wiki/Differential_calculus#Derivative>`_
of the quaternions [#]_ follows the formal definition:

.. math::
    \\begin{array}{rcl}
    \\dot{\\mathbf{q}}
    &=& \\underset{\\Delta t\\to 0}{\\mathrm{lim}} \\frac{\\mathbf{q}(t+\\Delta t)-\\mathbf{q}(t)}{\\Delta t} \\\\
    &=& \\underset{\\Delta t\\to 0}{\\mathrm{lim}} \\frac{1}{\\Delta t}\\Big(-2\\sin^2\\frac{\\|\\boldsymbol\\omega\\|\\Delta t}{4} + \\frac{\\boldsymbol\\omega}{\\|\\boldsymbol\\omega\\|}\\sin\\frac{\\|\\boldsymbol\\omega\\|\\Delta t}{2}\\Big)\\mathbf{q} \\\\
    &=& \\frac{\\boldsymbol\\omega}{\\|\\boldsymbol\\omega\\|} \\underset{\\Delta t\\to 0}{\\mathrm{lim}} \\frac{1}{\\Delta t}\\sin\\big(\\frac{\\|\\boldsymbol\\omega\\|\\Delta t}{2}\\big) \\mathbf{q} \\\\
    &=& \\frac{\\boldsymbol\\omega}{\\|\\boldsymbol\\omega\\|} \\frac{d}{dt} \\sin\\big(\\frac{\\|\\boldsymbol\\omega\\|}{2}t\\big) \\Big | _{t=0} \\; \\mathbf{q} \\\\
    &=& \\frac{1}{2}\\boldsymbol\\omega \\mathbf{q} \\\\
    &=& \\frac{1}{2}
    \\begin{bmatrix}
    -\\omega_x q_x -\\omega_y q_y - \\omega_z q_z\\\\
    \\omega_x q_w + \\omega_z q_y - \\omega_y q_z\\\\
    \\omega_y q_w - \\omega_z q_x + \\omega_x q_z\\\\
    \\omega_z q_w + \\omega_y q_x - \\omega_x q_y
    \\end{bmatrix}
    \\end{array}

.. warning::
    The product between the angular velocity :math:`\\boldsymbol\\omega\\in\\mathbb{R}^3`
    and the quaternion :math:`\\mathbf{q}\\in\\mathbb{H}^4` is done by
    representing the former as a pure quaternion :math:`\\boldsymbol\\omega=
    \\begin{bmatrix}0 & \\omega_x & \\omega_y & \\omega_z\\end{bmatrix}\\in\\mathbb{H}^4`,
    so that a `Hamilton Product <https://en.wikipedia.org/wiki/Quaternion#Hamilton_product>`_
    can be applied.

Defining the omega operator as

.. math::
    \\boldsymbol\\Omega(\\boldsymbol\\omega) =
    \\begin{bmatrix}0 & -\\boldsymbol\\omega^T \\\\ \\boldsymbol\\omega & -\\lfloor\\boldsymbol\\omega\\rfloor_\\times\\end{bmatrix} =
    \\begin{bmatrix}
    0 & -\\omega_x & -\\omega_y & -\\omega_z \\\\
    \\omega_x & 0 & \\omega_z & -\\omega_y \\\\
    \\omega_y & -\\omega_z & 0 & \\omega_x \\\\
    \\omega_z & \\omega_y & -\\omega_x & 0
    \\end{bmatrix}

.. note::
    The expression :math:`\\lfloor\\mathbf{x}\\rfloor_\\times` expands a vector
    :math:`\\mathbf{x}\\in\\mathbb{R}^3` into a :math:`3\\times 3`
    `skew-symmetric matrix <https://en.wikipedia.org/wiki/Skew-symmetric_matrix>`_:

    .. math::
        \\lfloor\\mathbf{x}\\rfloor_\\times =
        \\begin{bmatrix} 0 & -x_2 & x_1 \\\\ x_2 & 0 & -x_0 \\\\ -x_1 & x_0 & 0\\end{bmatrix}

we get an equivalent matrix expression for the derivative:

.. math::
    \\dot{\\mathbf{q}} = \\frac{1}{2}\\boldsymbol\\Omega(\\boldsymbol\\omega)\\mathbf{q}

This definition does **not** require a Hamilton product and the Omega operator
can be used as a common matrix to multiply with the quaternion as any other
linear operation.

Quaternion Integration
----------------------

First, let's remember the definition of a vector or `matrix exponential
<https://en.wikipedia.org/wiki/Matrix_exponential>`_ as a power series:

.. math::
    e^\\mathbf{X} = \\sum_{k=0}^\\infty \\frac{1}{k!} \\mathbf{X}^k

Letting :math:`\\mathbf{v}=\\mathbf{u}\\theta` be the **rotation vector**,
representing a rotation of :math:`\\theta` radians around the unitary axis
:math:`\\mathbf{u}=\\begin{bmatrix}u_x & u_y & u_z\\end{bmatrix}^T`, we can
get its exponential series as:

.. math::
    e^\\mathbf{v} = e^{\\mathbf{u}\\theta} = \\Big(1 - \\frac{\\theta^2}{2!} + \\frac{\\theta^4}{4!} + \\cdots\\Big)
    + \\Big(\\mathbf{u}\\theta - \\frac{\\mathbf{u}\\theta^3}{3!} + \\frac{\\mathbf{u}\\theta^5}{5!} + \\cdots\\Big)

We recognize the power-series expansion of `Euler's formula
<https://en.wikipedia.org/wiki/Euler%27s_formula#Using_power_series>`_, which
helps to map the quaternion :math:`\\mathbf{q}` from a rotation vector
:math:`\\mathbf{v}`. This **exponential map** :cite:p:`sola2017quaternion` is formerly defined as:

.. math::
    \\mathbf{q} = e^\\mathbf{v} =
    \\begin{bmatrix}\\cos\\frac{\\theta}{2} \\\\
    \\mathbf{u}\\sin\\frac{\\theta}{2}\\end{bmatrix}

Assuming the gyroscope data is sampled at a fixed rate and that the angular
velocity vector, :math:`\\boldsymbol\\omega`, in local body coordinates is
constant over the sampling interval :math:`\\Delta t`, we can alternatively
integrate :math:`\\dot{\\mathbf{q}}` to obtain:

.. math::
    \\mathbf{q}_{t+1} = e^{\\frac{1}{2}\\boldsymbol\\Omega(\\boldsymbol\\omega)\\Delta t}\\mathbf{q}_t

Using the `Euler-Rodrigues rotation formula
<https://en.wikipedia.org/wiki/Quaternions_and_spatial_rotation#Using_quaternion_as_rotations>`_
and the exponential map from above we find a **closed-form solution**
:cite:p:`spence1978`:

.. math::
    \\boxed{\\mathbf{q}_{t+1} =
    \\Bigg[
    \\cos\\Big(\\frac{\\|\\boldsymbol\\omega\\|\\Delta t}{2}\\Big)\\mathbf{I}_4 +
    \\frac{1}{\\|\\boldsymbol\\omega\\|}\\sin\\Big(\\frac{\\|\\boldsymbol\\omega\\|\\Delta t}{2}\\Big)\\boldsymbol\\Omega(\\boldsymbol\\omega)
    \\Bigg]\\mathbf{q}_t}

where :math:`\\mathbf{I}_4` is a :math:`4\\times 4` Identity matrix. The large
term inside the brackets, multiplying :math:`\\mathbf{q}_t`, is an orthogonal
rotation retaining the normalization of the propagated attitude quaternions.
Thus, it is not necessary to normalize :math:`\\mathbf{q}_{t+1}`, but it is
highly recommended to do so in order to avoid any `round-off errors
<https://en.wikipedia.org/wiki/Round-off_error>`_ inherent to all computers.

We can stop here and use this closed-form solution to update the quaternion,
but it cannot be used in a linear system. For that, we need to linearize the
quaternion update equation.

Quaternion Linearization
------------------------

Now we develop an :math:`n^{th}`-order polynomial linearization method built
from `Taylor series <https://en.wikipedia.org/wiki/Taylor_series>`_ of
:math:`\\mathbf{q}(t+\\Delta t)` around the time :math:`t` for the quaternion:

.. math::
    \\mathbf{q}_{t+1} = \\mathbf{q}_t + \\dot{\\mathbf{q}}_t\\Delta t +
    \\frac{1}{2!}\\ddot{\\mathbf{q}}_t\\Delta t^2 +
    \\frac{1}{3!}\\dddot{\\mathbf{q}}_t\\Delta t^3 + \\cdots

Using the definition of :math:`\\dot{\\mathbf{q}}` the new orientation
:math:`\\mathbf{q}_{t+1}` is written as :cite:p:`spence1978`:

.. math::
    \\begin{array}{rcl}
    \\mathbf{q}_{t+1} &=& \\Bigg[\\mathbf{I}_4 + \\frac{1}{2}\\boldsymbol\\Omega(\\boldsymbol\\omega)\\Delta t +
    \\frac{1}{2!}\\Big(\\frac{1}{2}\\boldsymbol\\Omega(\\boldsymbol\\omega)\\Delta t\\Big)^2 +
    \\frac{1}{3!}\\Big(\\frac{1}{2}\\boldsymbol\\Omega(\\boldsymbol\\omega)\\Delta t\\Big)^3 + \\cdots\\Bigg]\\mathbf{q}_t \\\\
    && \\qquad{} + \\frac{1}{4}\\dot{\\boldsymbol\\Omega}(\\boldsymbol\\omega)\\Delta t^2\\mathbf{q}_t
    + \\Big[\\frac{1}{12}\\dot{\\boldsymbol\\Omega}(\\boldsymbol\\omega)\\boldsymbol\\Omega(\\boldsymbol\\omega)
    + \\frac{1}{24}\\boldsymbol\\Omega(\\boldsymbol\\omega)\\dot{\\boldsymbol\\Omega}(\\boldsymbol\\omega)\\Big]\\Delta t^3\\mathbf{q}_t
    + \\frac{1}{12}\\ddot{\\boldsymbol\\Omega}(\\boldsymbol\\omega)\\Delta t^3\\mathbf{q}_t
    + \\cdots
    \\end{array}

Assuming the angular rate is constant over the period :math:`[t, t+1]`, we have
:math:`\\dot{\\boldsymbol\\omega}=0`, and we can prescind from the derivatives
of :math:`\\boldsymbol\\Omega` reducing the series to:

.. math::
    \\mathbf{q}_{t+1} = \\Bigg[\\mathbf{I}_4 + \\frac{1}{2}\\boldsymbol\\Omega(\\boldsymbol\\omega)\\Delta t +
    \\frac{1}{2!}\\Big(\\frac{1}{2}\\boldsymbol\\Omega(\\boldsymbol\\omega)\\Delta t\\Big)^2 +
    \\frac{1}{3!}\\Big(\\frac{1}{2}\\boldsymbol\\Omega(\\boldsymbol\\omega)\\Delta t\\Big)^3 + \\cdots\\Bigg]\\mathbf{q}_t

Notice the series for :math:`\\mathbf{q}_{t+1}` also follows the form of the
matrix exponential:

.. math::
    e^{\\frac{1}{2}\\boldsymbol\\Omega(\\boldsymbol\\omega)\\Delta t} =
    \\sum_{k=0}^\\infty \\frac{1}{k!} \\Big(\\frac{1}{2}\\boldsymbol\\Omega(\\boldsymbol\\omega)\\Delta t\\Big)^k

The error of the approximation vanishes rapidly at higher orders (:math:`k \\to
0`), or when the time step :math:`\\Delta t \\to 0`. The more terms we have,
the better our approximation should be, assuming the sensor signals are
unbiased and noiseless, with the downside of a big computational demand.

For our purpose a truncation up to the second term, making it of first order
(:math:`k=1`), is implemented.

.. math::
    \\boxed{\\mathbf{q}_{t+1} = \\Bigg[\\mathbf{I}_4 + \\frac{1}{2}\\boldsymbol\\Omega(\\boldsymbol\\omega)\\Delta t\\Bigg]\\mathbf{q}_t}

Two good reasons for this truncation are:

- Simple architectures (like embedded systems) avoid the burden of the heavy
  computation, and still achieve fairly good results.
- If the sensor signals are not properly filtered and corrected, the estimation
  will not converge to a good result, and even worsening it at higher orders.

The resulting quaternion must be normalized to operate as a versor, and to
represent a valid orientation.

.. math::
    \\mathbf{q}_{t+1} \\gets \\frac{\\mathbf{q}_{t+1}}{\\|\\mathbf{q}_{t+1}\\|}

Numerical Integration based on Runge-Kutta methods can be employed to increase
the accuracy, and are shown to be more effective. See :cite:p:`sola2017quaternion`
and :cite:p:`zhao2013` for a comparison of the different methods, their
accuracy, and their computational load.

Footnotes
---------

.. [#] A vector :math:`\\mathbf{x}` is unitary if its norm is equal to one:
    :math:`\\|\\mathbf{x}\\|=1`.
.. [#] The successive derivatives of :math:`\\mathbf{q}_n` are obtained by
    repeatedly applying the expression of the quaternion derivative,
    with :math:`\\ddot{\\boldsymbol\\omega}=0`.

    .. math::
        \\begin{array}{rcl}
        \\dot{\\mathbf{q}}_n &=& \\frac{1}{2}\\mathbf{q}_n\\boldsymbol\\omega_n \\\\
        \\ddot{\\mathbf{q}}_n &=& \\frac{1}{4}\\mathbf{q}_n\\boldsymbol\\omega^2_n + \\frac{1}{2}\\mathbf{q}_n\\dot{\\boldsymbol\\omega} \\\\
        \\dddot{\\mathbf{q}}_n &=& \\frac{1}{2^3}\\mathbf{q}_n\\boldsymbol\\omega^3_n + \\frac{1}{4}\\mathbf{q}_n\\dot{\\boldsymbol\\omega}\\boldsymbol\\omega_n + \\frac{1}{2}\\mathbf{q}\\boldsymbol\\omega_n\\dot{\\boldsymbol\\omega} \\\\
        \\mathbf{q}^{i\\geq 4}_n &=& \\frac{1}{2^i}\\mathbf{q}_n\\boldsymbol\\omega^i_n + \\cdots
        \\end{array}

    where all products and the powers of :math:`\\boldsymbol\\omega` are
    interpreted in terms of the quaternion product.

"""

from math import factorial
import numpy as np
from ..utils.core import _assert_numerical_iterable
from ..utils.core import _assert_numerical_positive_variable
from ..common.quaternion import Quaternion
from ..common.quaternion import QuaternionArray

class AngularRate:
    """
    Quaternion update by integrating measured angular velocities.

    All estimated quaternions are stored in the attribute ``Q`` as a
    :class:`numpy.ndarray` of shape *(N, 4)*, where *N* is the number of
    estimated attitudes. The first row is the initial attitude, and the last
    row is the final attitude.

    Parameters
    ----------
    gyr : numpy.ndarray, default: None
        N-by-3 array with measurements of angular velocity in rad/s.
    q0 : numpy.ndarray
        Initial orientation, as a versor (normalized quaternion).
    frequency : float, default: 100.0
        Sampling frequency in Herz.
    Dt : float, default: 0.01
        Sampling step in seconds. Inverse of sampling frequency. Not required
        if ``frequency`` value is given.
    method : str, default: ``'closed'``
        Estimation method to use. Options are: ``'series'``, ``'closed'``, or
        ``'integration'``. The option ``'integration'`` is a vectorized
        numerical integration of the angular velocity as roll-pitch-yaw angles.
    order : int
        Truncation order, if method ``'series'`` is used.
    representation : str, default: ``'quaternion'``
        Attitude representation. Options are ``'quaternion'``, ``'angles'`` or
        ``'rotmat'``.

    Examples
    --------
    >>> gyro_data.shape             # NumPy arrays with gyroscope data in rad/s
    (1000, 3)
    >>> from ahrs.filters import AngularRate
    >>> angular_rate = AngularRate(gyr=gyro_data)
    >>> angular_rate.Q
    array([[ 1.00000000e+00,  0.00000000e+00,  0.00000000e+00,  0.00000000e+00],
           [ 9.99999993e-01,  2.36511228e-06, -1.12991334e-04,  4.28771947e-05],
           [ 9.99999967e-01,  1.77775173e-05, -2.43529706e-04,  8.33144162e-05],
           ...,
           [-0.92576208, -0.23633121,  0.19738534, -0.2194337 ],
           [-0.92547793, -0.23388968,  0.19889139, -0.22187479],
           [-0.92504595, -0.23174096,  0.20086376, -0.22414251]])
    >>> angular_rate.Q.shape        # Estimated attitudes as Quaternions
    (1000, 4)

    The estimation of each attitude is built upon the previous attitude. This
    estimator sets the initial attitude equal to the unit quaternion
    ``[1.0, 0.0, 0.0, 0.0]`` by default, because we cannot obtain the first
    orientation with gyroscopes only.

    **Initial Values**

    We can use the class :class:`.Tilt` to estimate the initial attitude with a
    simple measurement of a tri-axial accelerometer:

    >>> from ahrs.filters import Tilt
    >>> tilt = Tilt()
    >>> q_initial = tilt.estimate(acc=acc_sample)  # One tridimensional sample suffices
    >>> angular_rate = AngularRate(gyr=gyro_data, q0=q_initial)
    >>> angular_rate.Q
    array([[ 0.77547502,  0.6312126 ,  0.01121595, -0.00912944],
           [ 0.77547518,  0.63121388,  0.01110125, -0.00916754],
           [ 0.77546726,  0.63122508,  0.01097435, -0.00921875],
           ...,
           [-0.92576208, -0.23633121,  0.19738534, -0.2194337 ],
           [-0.92547793, -0.23388968,  0.19889139, -0.22187479],
           [-0.92504595, -0.23174096,  0.20086376, -0.22414251]])

    :class:`.Tilt` can also use a magnetometer to improve the estimation
    with the heading orientation.

    >>> q_initial = tilt.estimate(acc=acc_sample, mag=mag_sample)
    >>> angular_rate = AngularRate(gyr=gyro_data, q0=q_initial)
    >>> angular_rate.Q
    array([[ 0.66475674,  0.55050651, -0.30902706, -0.39942875],
           [ 0.66473764,  0.5504497 , -0.30912672, -0.39946172],
           [ 0.66470495,  0.55039529, -0.30924191, -0.39950193],
           ...,
           [-0.90988476, -0.10433118,  0.28970402,  0.27802214],
           [-0.91087203, -0.1014633 ,  0.28977124,  0.2757716 ],
           [-0.91164416, -0.09861271,  0.2903888 ,  0.27359606]])

    """
    def __init__(self, gyr: np.ndarray = None, q0: np.ndarray = None, frequency: float = 100.0, order: int = 1, **kw):
        self.gyr: np.ndarray = gyr
        self.q0: np.ndarray = q0
        self.frequency: float = frequency
        self.order: int = order
        self.method: str = kw.get('method', 'closed')
        self.representation: str = kw.get('representation', 'quaternion')
        self.Dt: float = kw.get('Dt', 1.0/self.frequency if self.frequency else 0.01)
        # Check if inputs are valid
        self._assert_validity_of_inputs()
        if self.gyr is not None:
            _assert_numerical_iterable(self.gyr, 'gyr')
            if self.representation.lower() == 'quaternion':
                self.Q = self._compute_all()
            if self.representation.lower() == 'rotmat':
                self.R = self._compute_all()
            if self.representation.lower() == 'angles':
                self.W = self._compute_all()

    def _assert_validity_of_inputs(self):
        for item in ['frequency', 'Dt']:
            _assert_numerical_positive_variable(getattr(self, item), item)
        self.q0 = Quaternion(self.q0).to_array() if self.q0 is not None else np.array([1.0, 0.0, 0.0, 0.0])

    def _compute_all(self):
        """Estimate all quaternions with given sensor values."""
        self.gyr = np.copy(self.gyr)
        if self.gyr.ndim != 2:
            raise ValueError(f"Input 'gyr' must be a 2D array. Got {self.gyr.ndim}D instead.")
        # Compute with vectorized integration
        if self.method.lower() == 'integration':
            return self.integrate_angular_positions(self.gyr, self.Dt, representation=self.representation)
        # Compute with loop using method 'closed' or 'series'
        num_samples = self.gyr.shape[0]
        Q = np.zeros((num_samples, 4))
        Q[0] = self.q0
        for t in range(1, num_samples):
            Q[t] = self.update(Q[t-1], self.gyr[t], method=self.method, order=self.order)
        Q = QuaternionArray(Q)
        if self.representation.lower() == 'rotmat':
            return Q.to_DCM()
        if self.representation.lower() == 'angles':
            return Q.to_angles()
        return Q

    def integrate_angular_positions(self, gyr: np.ndarray, dt: float, representation: str = 'angles') -> np.ndarray:
        """
        Integrate angular positions :math:`\\mathbf{\\theta}` from
        instantaneous angular rates :math:`\\mathbf{\\omega}` at a time
        :math:`t` with a given time step :math:`\\Delta t`:

        .. math::
            \\mathbf{\\theta}_{t+1} = \\mathbf{\\theta}_t + \\mathbf{\\omega}_t \\Delta t

        Given the three main roll-pitch-yaw anglular velocities, it integrates
        them numerically with a simple `Riemann sum
        <https://en.wikipedia.org/wiki/Riemann_sum>`_, and returns the
        tri-axial angular positions.

        Calling this method is equivalent to::

            >>> angular_positions = numpy.cumsum(gyr * dt, axis=0)

        This method does not play a central role in this estimator, but can be
        used to obtain another reference to compare the results to.

        Parameters
        ----------
        gyr : array_like
            Angular rates, in rad/s.
        dt : float, optional
            Time step, in seconds. If not given, the time step is set to
            :math:`1/f_s`, where :math:`f_s` is the sampling frequency, which
            is defined as ``100`` Hz by default.
        representation : str, default: ``'angles'``
            Representation of the angular positions. Options are:
            ``'quaternion'``, ``'angles'`` or ``'rotmat'``.

        Returns
        -------
        theta : numpy.ndarray
            Tri-axial angular positions, in rad.
        """
        if representation.lower() not in ['quaternion', 'angles', 'rotmat']:
            raise ValueError(f"Representation must be 'quaternion', 'angles' or 'rotmat'. Got '{representation}' instead.")
        _assert_numerical_iterable(gyr, 'gyr')
        _assert_numerical_positive_variable(dt, 'dt')
        # Angular velocity integration --> Angular position
        angular_positions = np.cumsum(gyr * dt, axis=0)
        if representation.lower() == 'angles':
            return angular_positions
        # Return angular positions in the desired representation
        quaternions = QuaternionArray(rpy=angular_positions)
        if representation.lower() == 'quaternion':
            return quaternions.to_array()
        if representation.lower() == 'rotmat':
            return quaternions.to_DCM()

    def update(self, q: np.ndarray, gyr: np.ndarray, method: str = 'closed', order: int = 1, dt: float = None) -> np.ndarray:
        """
        Update the quaternion estimation

        Estimate quaternion :math:`\\mathbf{q}_{t+1}` from given a-priori
        quaternion :math:`\\mathbf{q}_t` with a measured instantaneous angular
        rate :math:`\\mathbf{\\omega}`.

        If ``method`` is ``'closed'``, the new orientation is computed with the
        closed-form solution:

        .. math::
            \\mathbf{q}_{t+1} =
            \\Bigg[
            \\cos\\Big(\\frac{\\|\\boldsymbol\\omega\\|\\Delta t}{2}\\Big)\\mathbf{I}_4 +
            \\frac{1}{\\|\\boldsymbol\\omega\\|}\\sin\\Big(\\frac{\\|\\boldsymbol\\omega\\|\\Delta t}{2}\\Big)\\boldsymbol\\Omega(\\boldsymbol\\omega)
            \\Bigg]\\mathbf{q}_t

        If ``method`` is ``'series'``, it is computed with a series of the form:

        .. math::
            \\mathbf{q}_{t+1} =
            \\Bigg[\\sum_{k=0}^\\infty \\frac{1}{k!} \\Big(\\frac{\\Delta t}{2}\\boldsymbol\\Omega(\\boldsymbol\\omega)\\Big)^k\\Bigg]\\mathbf{q}_t

        where the order :math:`k` in the series must be a non-negative integer.
        By default ``order`` is set to ``1``.

        Parameters
        ----------
        q : numpy.ndarray
            A-priori quaternion.
        gyr : numpy.ndarray
            Array with triaxial measurements of angular velocity in rad/s.
        method : str, default: ``'closed'``
            Estimation method to use. Options are: ``'series'`` or ``'closed'``.
        order : int
            Truncation order, if method ``'series'`` is used.
        dt : float, default: None
            Time step, in seconds, between consecutive Quaternions.

        Returns
        -------
        q : numpy.ndarray
            Estimated quaternion.

        Examples
        --------
        >>> from ahrs.filters import AngularRate
        >>> gyro_data.shape
        (1000, 3)
        >>> num_samples = gyro_data.shape[0]
        >>> Q = np.zeros((num_samples, 4))      # Allocation of quaternions
        >>> Q[0] = [1.0, 0.0, 0.0, 0.0]         # Initial attitude as a quaternion
        >>> angular_rate = AngularRate()
        >>> for t in range(1, num_samples):
        ...     Q[t] = angular_rate.update(Q[t-1], gyro_data[t])
        ...
        >>> Q
        array([[ 1.00000000e+00,  0.00000000e+00,  0.00000000e+00,  0.00000000e+00],
               [ 9.99999993e-01,  2.36511228e-06, -1.12991334e-04,  4.28771947e-05],
               [ 9.99999967e-01,  1.77775173e-05, -2.43529706e-04,  8.33144162e-05],
               ...,
               [-0.92576208, -0.23633121,  0.19738534, -0.2194337 ],
               [-0.92547793, -0.23388968,  0.19889139, -0.22187479],
               [-0.92504595, -0.23174096,  0.20086376, -0.22414251]])

        """
        _assert_numerical_iterable(q, 'q')
        _assert_numerical_iterable(gyr, 'gyr')
        if len(gyr) != 3:
            raise ValueError(f"gyr must be a 3-element array. Got {len(gyr)} instead.")
        if method.lower() not in ['series', 'closed']:
            raise ValueError(f"Invalid method '{method}'. Try 'series' or 'closed'")
        dt = self.Dt if dt is None else dt
        q = Quaternion(q)
        if np.linalg.norm(gyr) == 0:
            return q.to_array()
        Omega = np.array([
            [   0.0, -gyr[0], -gyr[1], -gyr[2]],
            [gyr[0],     0.0,  gyr[2], -gyr[1]],
            [gyr[1], -gyr[2],     0.0,  gyr[0]],
            [gyr[2],  gyr[1], -gyr[0],     0.0]])
        if method.lower() == 'closed':
            w = np.linalg.norm(gyr)
            A = np.cos(w*dt/2.0)*np.eye(4) + np.sin(w*dt/2.0)*Omega/w
        else:
            if order < 0:
                raise ValueError(f"The order must be an int equal to or greater than 0. Got {order}")
            S = 0.5 * dt * Omega
            A = np.identity(4)
            for i in range(1, order+1):
                A += S**i / factorial(i)
        q_new = Quaternion(A @ q)
        return q_new.to_array()



# -*- coding: utf-8 -*-
"""
Roberto Valenti's Algebraic Quaterion Algorithm (AQUA) :cite:p:`valenti2015`
estimates a quaternion with the algebraic solution of a system from
inertial+magnetic observations, solving `Wahba's Problem
<https://en.wikipedia.org/wiki/Wahba%27s_problem>`_.

AQUA computes the "tilt" quaternion and the "heading" quaternion separately in
two sub-parts. This avoids the impact of the magnetic disturbances on the roll
and pitch components of the orientation.

AQUA can be used with a complementary filter to fuse the gyroscope data
together with accelerometer and magnetic field readings. The correction part of
the filter is based on the independently estimated quaternions and works for
both IMU (Inertial Measurement Unit) and MARG (Magnetic, Angular Rate, and
Gravity) sensors :cite:p:`valenti2016`.

Quaternion as Orientation
-------------------------

Any orientation in a three-dimensional euclidean space of a frame :math:`A`
with respect to a frame :math:`B` can be represented by a unit quaternion
(a.k.a. versor), :math:`\\mathbf{q}\\in\\mathbb{H}^4`, in Hamiltonian space
defined as:

.. math::
    ^B_A\\mathbf{q} = \\begin{bmatrix}q_w\\\\q_x\\\\q_y\\\\q_z\\end{bmatrix} =
    \\begin{bmatrix}
    \\cos\\frac{\\alpha}{2}\\\\e_x\\sin\\frac{\\alpha}{2}\\\\e_y\\sin\\frac{\\alpha}{2}\\\\e_z\\sin\\frac{\\alpha}{2}
    \\end{bmatrix}

where :math:`\\alpha` is the rotation angle and
:math:`\\mathbf{e}=\\begin{bmatrix}e_x&e_y&e_z\\end{bmatrix}^T` is the unit
vector representing the rotation axis. The conjugate quaternion is used to
represent the orientation of frame :math:`B` relative to frame :math:`A`:

.. math::
    ^B_A\\mathbf{q}^* = \\,^A_B\\mathbf{q} = \\begin{bmatrix}q_w\\\\-q_x\\\\-q_y\\\\-q_z\\end{bmatrix}

The final orientation quaternion can be easily found through quaternion
multiplication:

.. math::
    ^C_A\\mathbf{q} = \\,^C_B\\mathbf{q} \\, ^B_A\\mathbf{q}

where the quaternion multiplication of any given quaternions :math:`\\mathbf{p}`
and :math:`\\mathbf{q}` is computed as:

.. math::
    \\mathbf{pq} =
    \\begin{bmatrix}
        p_w q_w - p_x q_x - p_y q_y - p_z q_z \\\\
        p_w q_x + p_x q_w + p_y q_z - p_z q_y \\\\
        p_w q_y - p_x q_z + p_y q_w + p_z q_x \\\\
        p_w q_z + p_x q_y - p_y q_x + p_z q_w
    \\end{bmatrix}

Unit quaternions in Hamiltonian spaces are often used to operate rotations of
vectors in a 3D euclidean space. A vector
:math:`^A\\mathbf{v}=\\begin{bmatrix}v_x & v_y & v_z\\end{bmatrix}^T`
expressed with respect to frame :math:`A` can be represented with respect to
frame :math:`B` as:

.. math::
    ^B\\mathbf{v}_q = \\, ^B_A\\mathbf{q} \\, ^A\\mathbf{v}_q \\, ^B_A\\mathbf{q}^*

where the vector :math:`^A\\mathbf{v}` is rewritten as the pure quaternion
:math:`^A\\mathbf{v}_q=\\begin{bmatrix}0 & v_x & v_y & v_z\\end{bmatrix}^T`, so
that it can multiply with the versor and its conjugate.

The inverse rotation can be achieved by flipping the versors:

.. math::
    ^A\\mathbf{v}_q = \\, ^B_A\\mathbf{q}^* \\, ^B\\mathbf{v}_q \\, ^B_A\\mathbf{q}
    = \\, ^A_B\\mathbf{q} \\, ^B\\mathbf{v}_q \\, ^A_B\\mathbf{q}^*

**Quaternion as Rotation Matrix in SO(3)**

These rotations can also be expressed within the three-dimensional euclidean
space if we express the rotation as a Direction Cosine Matrix
:math:`\\mathbf{R}\\in SO(3)`:

.. math::
    \\mathbf{R}(^B_A\\mathbf{q}) =
    \\begin{bmatrix}
    q_w^2+q_x^2-q_y^2-q_z^2 & 2(q_xq_y - q_wq_z) & 2(q_xq_z + q_wq_y) \\\\
    2(q_xq_y + q_wq_z) & q_w^2-q_x^2+q_y^2-q_z^2 & 2(q_yq_z - q_wq_x) \\\\
    2(q_xq_z - q_wq_y) & 2(q_wq_x + q_yq_z) & q_w^2-q_x^2-q_y^2+q_z^2
    \\end{bmatrix}

And, because the Direction Cosine Matrix belongs to the Special Orthogonal
Group :math:`SO(3)`, the inverse rotation is simply its transpose:

.. math::
    ^A\\mathbf{v} = \\mathbf{R}(^A_B\\mathbf{q})\\,^B\\mathbf{v} = \\mathbf{R}^T(^B_A\\mathbf{q})\\,^B\\mathbf{v}

.. note::
    For a more detailed explanation of quaternions and their use in spatial
    rotation see the documentation of the class `Quaternions <../quaternion.html>`_.

Given two quaternions :math:`\\mathbf{p}` and :math:`\\mathbf{q}`, the cosine
of the angle :math:`\\Omega` subtended by the arc between them is equal to the
dot product of the two quaternions.

.. math::
    \\cos\\Omega = \\mathbf{p}\\cdot\\mathbf{q} = p_wq_w + p_xq_x + p_yq_y + p_zq_z

It is easy to see that the dot product of any :math:`\\mathbf{q}=\\begin{bmatrix}q_w & q_x & q_y & q_z\\end{bmatrix}`
and the identity quaternion :math:`\\mathbf{q}_I=\\begin{bmatrix}1 & 0 & 0 & 0\\end{bmatrix}`
is equal to the :math:`q_w` component:

.. math::
    \\mathbf{q}\\cdot\\mathbf{q}_I = q_w

**Interpolations**

The simple **Linear intERPolation** (LERP) between two quaternions :math:`\\mathbf{p}`
and :math:`\\mathbf{q}` is obtained as:

.. math::
    \\overline{\\mathbf{r}} = (1-\\alpha)\\mathbf{p} + \\alpha\\mathbf{q}

where :math:`\\alpha\\in [0, 1]`. But this does not keep the unit norm, so we
must normalize the resulting interpolation:

.. math::
    \\widehat{\\mathbf{r}} = \\frac{\\overline{\\mathbf{r}}}{\\|\\overline{\\mathbf{r}}\\|}

The **Spherical Linear intERPolation** (SLERP) gives a correct evaluation of
the weighted average of two points lying on a curve. In the case of quaternions,
the points lie on the surface of the 4D sphere (hypersphere).

.. math::
    \\widehat{\\mathbf{r}} = \\frac{\\sin([1-\\alpha]\\Omega)}{\\sin\\Omega}\\mathbf{p} +
    \\frac{\\sin(\\alpha\\Omega)}{\\sin\\Omega}\\mathbf{q}

Quaternion from Earth-Field Observations
----------------------------------------

The local (sensor) frame is labeled as :math:`L`, and the global (Earth) frame
as :math:`G`. The *measured* acceleration, :math:`^L\\mathbf{a}`, and the true
Earth gravitational acceleration, :math:`^G\\mathbf{g}`, are defined as **unit
vectors** [#]_:

.. math::
    \\begin{array}{rcl}
    ^L\\mathbf{a} &=& \\begin{bmatrix}a_x & a_y & a_z\\end{bmatrix}^T \\\\ && \\\\
    ^G\\mathbf{g} &=& \\begin{bmatrix}0 & 0 & 1\\end{bmatrix}^T
    \\end{array}

The *measured* local magnetic field, :math:`^L\\mathbf{m}`, and the true
geomagnetic field, :math:`^G\\mathbf{h}`, are also unit vectors:

.. math::
    \\begin{array}{rcl}
    ^L\\mathbf{m} &=& \\begin{bmatrix}m_x & m_y & m_z\\end{bmatrix}^T \\\\ && \\\\
    ^G\\mathbf{h} &=& \\begin{bmatrix}h_x & h_y & h_z\\end{bmatrix}^T
    \\end{array}

The gyroscopes measure the angular velocity, :math:`^L\\mathbf{\\omega}`,
around the three sensor frame axes:

.. math::
    ^L\\mathbf{\\omega} = \\begin{bmatrix}\\omega_x & \\omega_y & \\omega_z\\end{bmatrix}^T

The measured angular velocities are **not** normalized, unlike the other
sensors, and are assumed to be in *radians per second*.

A straightforward way to formulate the quaternion, :math:`^L_G\\mathbf{q}`,
relating the global frame :math:`G` to the local frame :math:`L`, is through
the inverse orientation which rotates the measured quantities :math:`^L\\mathbf{a}`
and :math:`^L\\mathbf{m}` into the reference quantities :math:`^G\\mathbf{g}`
and :math:`^G\\mathbf{h}`:

.. math::
    \\begin{array}{rcl}
    \\mathbf{R}^T(^L_G\\mathbf{q})\\,^L\\mathbf{a} &=& \\,^G\\mathbf{g} \\\\ && \\\\
    \\mathbf{R}^T(^L_G\\mathbf{q})\\,^L\\mathbf{m} &=& \\,^G\\mathbf{h}
    \\end{array}

In the case of a disagreement between the gravitational and magnetometer
readings, the system will not have a solution. To address this problem a
modified equation system is built.

First, the global coordinate frame :math:`G` is aligned with the magnetic North.
The global frame's X-axis points to the same direction as the local magnetic
field (Z-axis remains vertical.)

.. warning::
    This global frame is only *fixed* in case the local magnetic field does not
    change its heading. Thus, no magnetic inference should be present.

Let :math:`^G\\Pi_{zx^+}` be the half-plane which contains all points that lie
in the global XZ-plane such that X is non-negative.

The magnetic reading, when rotated into the global frame, must lie on the
half-plane :math:`^G\\Pi_{zx^+}` to guarantee that the heading will be measured
with respect to the geomagnetic North.

.. math::
    \\begin{array}{lc}
    \\mathbf{R}^T(^L_G\\mathbf{q})\\,^L\\mathbf{a} = \\,^G\\mathbf{g} &\\\\ & \\\\
    \\mathbf{R}^T(^L_G\\mathbf{q})\\,^L\\mathbf{m} \\in \\,^G\\Pi_{zx^+} &
    \\end{array}

This way we don't need a-priori knowledge of the direction of Earth's magnetic
field :math:`^G\\mathbf{h}`.

The orientation :math:`^L_G\\mathbf{q}` is decomposed into two auxiliary
quaternions, such that:

.. math::
    \\begin{array}{rcl}
    ^L_G\\mathbf{q} &=& \\mathbf{q}_\\mathrm{acc} \\mathbf{q}_\\mathrm{mag} \\\\ && \\\\
    \\mathbf{R}(^L_G\\mathbf{q}) &=& \\mathbf{R}(\\mathbf{q}_\\mathrm{acc}) \\mathbf{R}(\\mathbf{q}_{\\mathrm{mag}})
    \\end{array}

The quaternion :math:`\\mathbf{q}_{\\mathrm{mag}}` represents a rotation around
the Z-axis to point North only:

.. math::
    \\mathbf{q}_{\\mathrm{mag}} = \\begin{bmatrix}q_{w\\mathrm{mag}} & 0 & 0 & q_{z\\mathrm{mag}}\\end{bmatrix}^T

Quaternion from Accelerometer
-----------------------------

The auxiliary quaternion :math:`\\mathbf{q}_\\mathrm{acc}` is built as a
function of :math:`^L\\mathbf{a}`. The observations of the gravity vector in
the two reference frames allows us to find the quaternion that performs the
transformation between the two representations.

.. math::
    \\begin{array}{rcl}
    \\mathbf{R}(^L_G\\mathbf{q})\\,^G\\mathbf{g} &=& \\, ^L\\mathbf{a} \\\\
    \\mathbf{R}(\\mathbf{q}_\\mathrm{acc})\\mathbf{R}(\\mathbf{q}_\\mathrm{mag})
    \\begin{bmatrix}0 \\\\ 0 \\\\ 1\\end{bmatrix} &=&
    \\begin{bmatrix}a_x \\\\ a_y \\\\ a_z\\end{bmatrix}
    \\end{array}

The representation of the gravity vector in the global frame :math:`G` only has
a component on the Z-axis. Thus, any rotation about this axis does not produce
any change on it.

.. math::
    \\mathbf{R}(\\mathbf{q}_\\mathrm{acc})
    \\begin{bmatrix}0 \\\\ 0 \\\\ 1\\end{bmatrix} =
    \\begin{bmatrix}a_x \\\\ a_y \\\\ a_z\\end{bmatrix}

The alignment of the gravity vector from global frame into local frame can be
achieved by infinite rotations with definite roll and pitch angles and
arbitrary yaw. To restrict the solutions to a finite number :math:`q_{z\\mathrm{acc}}=0`
is chosen.

This gives four solutions of :math:`\\mathbf{q}_\\mathrm{acc}`. Two are
discarded for having a negative norm, and from the remaining two the one with a
positive :math:`q_w` is taken.

.. math::
    \\mathbf{q}_\\mathrm{acc} =
    \\begin{bmatrix}\\sqrt{\\frac{a_z+1}{2}} & -\\frac{a_y}{\\sqrt{2(a_z+1)}} & \\frac{a_x}{\\sqrt{2(a_z+1)}} & 0\\end{bmatrix}^T

But it has a singularity at :math:`a_z=-1`. Therefore, the final solution has to
be defined as:

.. math::
    \\mathbf{q}_\\mathrm{acc} =
    \\left\\{
    \\begin{array}{ll}
        \\begin{bmatrix}\\sqrt{\\frac{a_z+1}{2}} & -\\frac{a_y}{\\sqrt{2(a_z+1)}} & \\frac{a_x}{\\sqrt{2(a_z+1)}} & 0\\end{bmatrix}^T & \\mathrm{if}\\; a_z \\geq 0 \\\\
        \\begin{bmatrix}-\\frac{a_y}{\\sqrt{2(1-a_z)}} & \\sqrt{\\frac{1-a_z}{2}} & 0 & \\frac{a_x}{\\sqrt{2(1-a_z)}}\\end{bmatrix}^T & \\mathrm{if}\\; a_z < 0
    \\end{array}
    \\right.

Quaternion from Magnetometer
----------------------------

The auxiliary quaternion :math:`\\mathbf{q}_\\mathrm{mag}` is derived as a
function of :math:`^L\\mathbf{m}` and :math:`\\mathbf{q}_\\mathrm{acc}`.

The quaternion :math:`\\mathbf{q}_\\mathrm{acc}` is used to rotate the magnetic
field vector :math:`^L\\mathbf{m}` into an intermediate frame whoe Z-axis is
the same as the global coordinate frame with X- and Y-axes pointing to unknown
directions due to the unknown yaw of :math:`\\mathbf{q}_\\mathrm{acc}`.

.. math::
    \\mathbf{R}^T(\\mathbf{q}_\\mathrm{acc})\\,^L\\mathbf{m} = \\mathbf{l}

where :math:`\\mathbf{l}` is the rotated magnetic field vector. Then, we find
the quaternion :math:\\mathbf{q}_\\mathrm{mag} that rotates the vector
:math:`\\mathbf{l}` into the vector that lies on :math:`^G\\Pi_{zx^+}`:

.. math::
    \\mathbf{R}^T(\\mathbf{q}_\\mathrm{mag})
    \\begin{bmatrix}l_x \\\\ l_y \\\\ l_z\\end{bmatrix} =
    \\begin{bmatrix}\\sqrt{\\Gamma} \\\\ 0 \\\\ l_z\\end{bmatrix}

where :math:`\\Gamma=l_x^2+l_y^2`. This quaternion performs a rotation only
about the global Z-axis to align the X-axis of the intermediate frame with the
X-axis of the global frame without affecting the roll and pitch. Therefore, if
there is a magnetic interference, it would affect only the headin angle.

The solution to find this quaternion ensuring the shortest rotation is:

.. math::
    \\mathbf{q}_\\mathrm{mag} =
    \\begin{bmatrix}\\frac{\\sqrt{\\Gamma + l_x\\sqrt{\\Gamma}}}{\\sqrt{2\\Gamma}} & 0 & 0 &\\frac{l_y}{\\sqrt{2}\\sqrt{\\Gamma+l_x\\sqrt{\\Gamma}}}\\end{bmatrix}^T

This quaternion has a singularity too, but here it happens when :math:`l_x<0`
and :math:`l_y=0`. Eventually, a simliar solution is found restraining the
condition of :math:`l_x`:

.. math::
    \\mathbf{q}_\\mathrm{mag} =
    \\left\\{
    \\begin{array}{ll}
        \\begin{bmatrix}\\frac{\\sqrt{\\Gamma + l_x\\sqrt{\\Gamma}}}{\\sqrt{2\\Gamma}} & 0 & 0 & \\frac{l_y}{\\sqrt{2}\\sqrt{\\Gamma+l_x\\sqrt{\\Gamma}}}\\end{bmatrix}^T & \\mathrm{if}\\; l_x \\geq 0 \\\\
        \\begin{bmatrix}\\frac{l_y}{\\sqrt{2}\\sqrt{\\Gamma-l_x\\sqrt{\\Gamma}}} & 0 & 0 & \\frac{\\sqrt{\\Gamma - l_x\\sqrt{\\Gamma}}}{\\sqrt{2\\Gamma}}\\end{bmatrix}^T & \\mathrm{if}\\; l_x < 0
    \\end{array}
    \\right.

The generalized quaternion orientation of the global frame relative to the
local frame as the mulitplication of two quaternions :math:`\\mathbf{q}_\\mathrm{acc}`
and :math:`\\mathbf{q}_\\mathrm{mag}`:

.. math::
    ^L_G\\mathbf{q} = \\mathbf{q}_\\mathrm{acc} \\, \\mathbf{q}_\\mathrm{mag}

The quaternion :math:`^L_G\\mathbf{q}` does not suffer from the discontinuity
problem of the yaw angle given by the switching formulation of
:math:`\\mathbf{q}_\\mathrm{acc}` thanks to the multiplication with
:math:`\\mathbf{q}_\\mathrm{mag}`, which performs the alignment of the
intermediate frame into the global frame.

Quaternion-Based Complementary Filter
-------------------------------------

A complementary filter fuses attitude estimation in quaternion form from
gyroscope data with accelerometer and magnetometer data in the form of a delta
quaternion.

If only IMU data is provided (gyroscopes and accelerometers), it corrects only
roll and pitch of the attitude. If magnetometer data is also provided (MARG) a
second step is added to the algorithm where a magnetic delta quaternion is
derived to correct the heading of the previous estimation by aligning the
current frame with the magnetic field.

In the **Prediction** step the measured angular velocity is used to compute a
first estimation of the orientation in quaternion form.

Most literature estimating the quaternion derivative from an angular rate
measurement is usually calculated for the one representing the orientation of
the local frame with respect to the global frame.

.. math::
    ^G_L\\dot{\\mathbf{q}}_{\\omega, t_k}=\\frac{1}{2}\\,^G_L\\mathbf{q}_{t_{k-1}}\\,^L\\mathbf{\\omega}_{q, t_k}

However, Valenti uses the inverse orientation, so the quaternion derivative is
computed using the inverse unit quaternion, which is simply the conjugate:

.. math::
    ^L_G\\dot{\\mathbf{q}}_{\\omega, t_k}= \\,^G_L\\dot{\\mathbf{q}}_{\\omega, t_k}^*=-\\frac{1}{2}\\,^L\\mathbf{\\omega}_{q, t_k}\\,^L_G\\mathbf{q}_{t_{k-1}}

where :math:`^L\\mathbf{\\omega}_{q, t_k}=\\begin{bmatrix}0 & \\omega_x & \\omega_y & \\omega_z\\end{bmatrix}^T`
is the measured angular velocity, in radians per second, arranged as a pure
quaternion at time :math:`t_k`, and :math:`^L_G\\mathbf{q}_{t_{k-1}}` is the
previous estimate of the orientation.

The orientation of the global frame relative to local frame at time :math:`t_k`
can be finally computed by numerically integrating the quaternion derivative
using the sampling period :math:`\\Delta t = t_k - t_{k-1}`.

.. math::
    ^L_G\\mathbf{q}_{\\omega, t_k} = \\, ^L_G\\mathbf{q}_{t_{k-1}} + \\,^L_G\\dot{\\mathbf{q}}_{\\omega, t_k}\\Delta t

The **Correction** step is based on a multiplicative approach, where the
predicted quaternion :math:`^L_G\\mathbf{q}_\\omega` is corrected by means of
two *delta quaternions*:

.. math::
    \\begin{equation}
    ^L_G\\mathbf{q} = \\, ^L_G\\mathbf{q}_\\omega \\, \\Delta\\mathbf{q}_\\mathrm{acc} \\, \\Delta\\mathbf{q}_\\mathrm{mag}
    \\end{equation}

The delta quaternions are computed and filtered independently by the
high-frequency noise. This correction is divided in two steps: correction of
roll and pitch of the predicted quaternion, and then the correction of the yaw
angle if readings of the magnetic field are provided.

Accelerometer-Based Correction
------------------------------

The inverse predicted quaternion :math:`^G_L\\mathbf{q}_\\omega` is used to
rotate the normalized body frame gravity vector :math:`^L\\mathbf{a}`,
*measured* by the accelerometer, into the global frame:

.. math::
    \\mathbf{R}(^G_L\\mathbf{q}_\\omega)\\,^L\\mathbf{a} = \\,^G\\mathbf{g}_p

where :math:`^G\\mathbf{g}_p=\\begin{bmatrix}g_x & g_y & g_z\\end{bmatrix}^T`
is the **predicted gravity**, which always has a small deviation from the real
gravity vector :math:`^G\\mathbf{g}=\\begin{bmatrix}0 & 0 & 1\\end{bmatrix}^T`.
We compute the delta quaternion :math:`\\Delta\\mathbf{q}_\\mathrm{acc}` to
rotate :math:`^G\\mathbf{g}` into :math:`^G\\mathbf{g}_p`:

.. math::
    \\mathbf{R}(\\Delta\\mathbf{q}_\\mathrm{acc}) \\, ^G\\mathbf{g} = \\,^G\\mathbf{g}_p

Similar to the auxiliary quaternions, we find a closed-form solution:

.. math::
    \\Delta\\mathbf{q}_\\mathrm{acc} =
    \\begin{bmatrix}\\sqrt{\\frac{g_z+1}{2}} & - \\frac{g_y}{\\sqrt{2(g_z+1)}} & \\frac{g_x}{\\sqrt{2(g_z+1)}} & 0\\end{bmatrix}^T

This has a singularity at :math:`g_z=-1`, but it can be ignored, because the
value of :math:`g_z` will always be closer to 1.

The delta quaternion is affected by the accelerometer's high frequency noise,
so we scale it down by using an interpolation with the identity quaternion
:math:`\\mathbf{q}_I`. As demonstrated above, the dot product with
:math:`\\mathbf{q}_I` is equal to the :math:`\\Delta q_{w\\mathrm{acc}}`
component of :math:`\\Delta\\mathbf{q}_\\mathrm{acc}`.

If :math:`\\Delta q_{w\\mathrm{acc}}>\\epsilon`, where :math:`\\epsilon` is a
threshold value (default is :math:`\\epsilon=0.9`), a simple LERP is used:

.. math::
    \\widehat{\\Delta\\mathbf{q}}_\\mathrm{acc} = \\frac{\\overline{\\Delta\\mathbf{q}}_\\mathrm{acc}}{\\|\\overline{\\Delta\\mathbf{q}}_\\mathrm{acc}\\|}

The predicted quaternion from gyroscopes is multiplied with the delta
quaternion to correct the roll and pitch components:

.. math::
    ^L_G\\mathbf{q}' = \\, ^L_G\\mathbf{q}_\\omega \\, \\widehat{\\Delta\\mathbf{q}}_\\mathrm{acc}

The *heading angle* predicted by the gyroscope integration is *not corrected*
in this step.

Magnetometer-Based Correction
-----------------------------

When a magnetic field measurement is provided the second step corrects the
heading component. We use the quaternion inverse of :math:`^L_G\\mathbf{q}'` to
rotate the magnetic field vector :math:`^L\\mathbf{m}` from the body frame into
the world frame.

.. math::
    \\mathbf{R}(^L_G\\mathbf{q}')\\,^L\\mathbf{m} = \\mathbf{l}

The delta quaternion :math:`\\Delta\\mathbf{q}_\\mathrm{mag}` rotates the
vector :math:`\\mathbf{l}` into the vector that lies on the XZ-semiplane:

.. math::
    \\mathbf{R}^T(\\Delta\\mathbf{q}_\\mathrm{mag})
    \\begin{bmatrix}l_x \\\\ l_y \\\\ l_z\\end{bmatrix} =
    \\begin{bmatrix}\\sqrt{l_x^2+l_y^2} \\\\ 0 \\\\ l_z\\end{bmatrix}

The solution to the above ensures the shortest rotation:

.. math::
    \\Delta\\mathbf{q}_\\mathrm{mag} = \\begin{bmatrix}
    \\frac{\\sqrt{\\Gamma+l_x\\sqrt{\\Gamma}}}{\\sqrt{2\\Gamma}} & 0 & 0 &
    \\frac{l_y}{\\sqrt{2(\\Gamma+l_x\\sqrt{\\Gamma})}}
    \\end{bmatrix}^T

This delta quaternion is affected by the noise of the magnetometer, which is
also filtered like the :math:`\\Delta\\mathbf{q}_\\mathrm{acc}` switching
between LERP and SLERP according to the same criterion.

Because :math:`\\Delta\\mathbf{q}_\\mathrm{acc}` and
:math:`\\Delta\\mathbf{q}_\\mathrm{mag}` are affected independently with
different noises, two different thresholds can be used: :math:`\\alpha` for the
accelerometer and :math:`\\beta` for the magnetometer.

Finally, :math:`\\widehat{\\Delta\\mathbf{q}}_\\mathrm{mag}` is multiplied with
:math:`^L_G\\mathbf{q}'` to obtain the orientation of the global frame with
respect to the local frame:

.. math::
    ^L_G\\mathbf{q} = \\,^L_G\\mathbf{q}' \\, \\widehat{\\Delta\\mathbf{q}}_\\mathrm{mag}

Adaptive Gain
-------------

When the vehicle moves with high acceleration, the magnitude and direction of
the total measured acceleration vector are different from gravity, and the
attitude is estimated using a false reference.

However, the gyroscope readings are not affected by linear acceleration, thus
they can still be used to compute a relatively accurate orientation estimation.

A fusion algorithm using a constant gain to decide between LERP and SLERP
cannot overcome the aforementioned problem if such gain has been defined for
static conditions.

An adaptive gain dependant on a dynamic state can be used to tackle this
problem.

First, a **magnitude error** :math:`e_m` is defined:

.. math::
    e_m = \\frac{|\\|\\,^L\\tilde{\\mathbf{a}}\\|-g|}{g}

where :math:`\\|\\,^L\\tilde{\\mathbf{a}}\\|` is the norm of the measured local
frame acceleration vector before normalization, and :math:`g=9.81 \\, \\frac{m}{s^2}`.

The magnitude error tells us how different the measured acceleration from the
gravitational acceleration is. When the magnitude error is zero, the
acceleration is only due to gravity and no extra acceleration force is present.

If the magnitude error is greater than zero, the acceleration is due to the
gravity PLUS an extra acceleration force.

We make the filtering gain :math:`\\alpha` dependent on the magnitude error
:math:`e_m` through the gain factor :math:`f`:

.. math::
    \\alpha = \\overline{\\alpha}f(e_m)

where :math:`\\overline{\\alpha}` is the constant value that gives the best
filtering result in static conditions and :math:`f(e_m)` is what is called the
**gain factor**, which is a piecewise continuous function of the magnitude
error.

The gain factor is equal to :math:`1` when the magnitude of the
non-gravitational acceleration is not high enough to overcome the acceleration
gravity and the value of the error magnitude does not reach the first threshold
:math:`t_1`.

.. math::
    f(e_m) =
    \\left\\{
    \\begin{array}{ll}
        1 & \\mathrm{if}\\; e_m \\leq t_1 \\\\
        \\frac{t_2-e_m}{t_1} & \\mathrm{if}\\; t_1 < e_m < t_2 \\\\
        0 & \\mathrm{if}\\; e_m \\geq t_2
    \\end{array}
    \\right.

If the non-gravitational acceleration rises and the error magnitude
exceeds that first threshold, the gain factor decreases linearly with the
increase of the magnitude error until reaching zero for error magnitude equal
to the second threshold :math:`t_2` and over.

Empirically, the authors found that the threshold values giving the best
results are :math:`0.1` and :math:`0.2`.

.. image:: ../images/adaptive_gain.png

In summary:

- If the system is static (:math:`e_m < t_1`) : :math:`f(e_m)=1` and :math:`\\alpha = \\overline{\\alpha}`
- If the system is in motion (:math:`t_1 < e_m < t_2`) : :math:`f(e_m) \\in (0, 1)` and :math:`\\alpha = \\overline{\\alpha}f(e_m)`
- If the system is in high acceleration (:math:`e_m > t_2`) : :math:`f(e_m)=0` and :math:`\\alpha = 0`

Filter Initialization
---------------------

The values of the current body-frame acceleration and magnetic field vectors
are used to produce the quaternion representing the initial orientation of the
rigid body in any configuration. Therefore, for the initialization, the filter
does not need any assumption and it is performed in one single step.

.. math::
    ^L_G\\mathbf{q}_0 = \\mathbf{q}_\\mathrm{acc} \\mathbf{q}_\\mathrm{mag}

The bias of a gyroscope's reading is a slow-varying signal considered as low
frequency noise. A low-pass filtering is applied to separate the bias from the
actual angular velocity, but only when the sensor is in a steady-state
condition to avoid filtering useful information.

If the sensor is in the steady-state condition, the bias is updated, otherwise
it is assumed to be equal to the previous step value. The estimated bias is
then subtracted from the gyroscope reading obtaining a bias-free angular
velocity measurement.

Footnotes
---------
.. [#] Any vector :math:`\\mathbf{x}` is a **unit vector** if :math:`\\|\\mathbf{x}\\|=1`.

"""

import numpy as np
from ..common.quaternion import Quaternion
from ..common.constants import MUNICH_LATITUDE, MUNICH_HEIGHT
from ..utils.core import _assert_numerical_iterable
from ..utils.core import _assert_same_shapes

# Reference Observations in Munich, Germany
from ..utils.wgs84 import WGS
GRAVITY = WGS().normal_gravity(MUNICH_LATITUDE, MUNICH_HEIGHT)

def slerp_I(q: np.ndarray, ratio: float, t: float) -> np.ndarray:
    """
    Interpolation with identity quaternion

    Interpolate a given quaternion with the identity quaternion
    :math:`\\mathbf{q}_I=\\begin{pmatrix}1 & 0 & 0 & 0\\end{pmatrix}` to
    scale it to closest versor.

    The interpolation can be with either LERP (Linear) or SLERP (Spherical
    Linear) methods, decided by a threshold value :math:`t`, which lies
    between ``0.0`` and ``1.0``.

    .. math::
        \\mathrm{method} = \\left\\{
        \\begin{array}{ll}
            \\mathrm{LERP} & \\: q_w > t \\\\
            \\mathrm{SLERP} & \\: \\mathrm{otherwise}
        \\end{array}
        \\right.

    For LERP, a simple equation is implemented:

    .. math::
        \\hat{\\mathbf{q}} = (1-\\alpha)\\mathbf{q}_I + \\alpha\\Delta \\mathbf{q}

    where :math:`\\alpha\\in [0, 1]` is the gain characterizing the cut-off
    frequency of the filter. It basically decides how "close" to the given
    quaternion or to the identity quaternion the interpolation is.

    If the scalar part :math:`q_w` of the given quaternion is below the
    threshold :math:`t`, SLERP is used:

    .. math::
        \\hat{\\mathbf{q}} = \\frac{\\sin([1-\\alpha]\\Omega)}{\\sin\\Omega} \\mathbf{q}_I + \\frac{\\sin(\\alpha\\Omega)}{\\sin\\Omega} \\mathbf{q}

    where :math:`\\Omega=\\arccos(q_w)` is the `subtended angle
    <https://en.wikipedia.org/wiki/Subtended_angle>`_ between the quaternions.

    Parameters
    ----------
    q : numpy.array
        Quaternion to inerpolate with.
    ratio : float
        Gain characterizing the cut-off frequency of the filter.
    t : float
        Threshold deciding interpolation method. LERP when qw>t, otherwise
        SLERP.

    Returns
    -------
    q : numpy.array
        Interpolated quaternion
    """
    q_I = np.array([1.0, 0.0, 0.0, 0.0])
    if q[0] > t:  # LERP
        q = (1.0-ratio)*q_I + ratio*q   # (eq. 50)
    else:       # SLERP
        angle = np.arccos(q[0])
        q = q_I*np.sin(abs(1.0-ratio)*angle)/np.sin(angle) + q*np.sin(ratio*angle)/np.sin(angle)    # (eq. 52)
    q /= np.linalg.norm(q)              # (eq. 51)
    return q

def adaptive_gain(a_local: np.ndarray, alpha_bar: float = 0.1, t1: float = 0.1, t2: float = 0.2, g: float = GRAVITY) -> float:
    """
    Adaptive filter gain factor

    The estimated gain :math:`\\alpha` is dependent on the gain factor
    :math:`f(e_m)`:

    .. math::
        \\alpha = \\overline{\\alpha} f(e_m)

    where the magnitude error is defined by the measured acceleration
    :math:`\\mathbf{a}` before normalization and the reference gravity
    :math:`g\\approx 9.809196 \\, \\frac{m}{s^2}`:

    .. math::
        e_m = \\frac{|\\|\\,^L\\tilde{\\mathbf{a}}\\|-g|}{g}

    The gain factor is constant and equal to 1 when the magnitude of the
    nongravitational acceleration is not high enough to overcome gravity.

    If nongravitational acceleration rises and :math:`e_m` exceeds the
    first threshold, the gain factor :math:`f` decreases linearly with the
    increase of the magnitude until reaching zero at the second threshold
    and above it.

    .. math::
        f(e_m) =
        \\left\\{
        \\begin{array}{ll}
            1 & \\mathrm{if}\\; e_m \\leq t_1 \\\\
            \\frac{t_2-e_m}{t_1} & \\mathrm{if}\\; t_1 < e_m < t_2 \\\\
            0 & \\mathrm{if}\\; e_m \\geq t_2
        \\end{array}
        \\right.

    Empirically, both thresholds have been defined at ``0.1`` and ``0.2``,
    respectively. They can be changed by setting the values of input parameters
    ``t1`` and ``t2``.

    Parameters
    ----------
    a_local : numpy.ndarray
        Measured local acceleration vector.
    alpha_bar : float, default: 0.1
        Gain yielding best results in static conditions.
    t1 : float, default: 0.1
        First threshold.
    t2 : float, default: 0.2
        Second threshold.
    g : float, default: 9.809196
        Reference gravitational acceleration in m/s^2. The estimated gravity in
        Munich, Germany (``9.809196``) is used as default reference value.

    Returns
    -------
    gain : float
        Gain factor.

    Examples
    --------
    >>> from ahrs.filters.aqua import adaptive_gain
    >>> acc = np.array([0.0699, 9.7688, -0.2589])   # Measured acceleration. Quasi-static state.
    >>> adaptive_gain(acc)
    0.1
    >>> acc = np.array([0.8868, 10.8803, -0.4562])  # New measured acceleration. Slightly above first threshold.
    >>> adaptive_gain(acc)
    0.08615664547367627
    >>> acc = np.array([4.0892, 12.7667, -2.6047])  # New measured acceleration. Above second threshold.
    >>> adaptive_gain(acc)
    0.0
    >>> adaptive_gain(acc, alpha_bar=0.01, t1=0.2, t2=0.5)   # Same acceleration. New thresholds.
    0.005377771414416655
    >>> adaptive_gain(acc, alpha_bar=0.01, t1=0.2, t2=0.5, g=9.82)   # Same acceleration and thresholds. New reference gravity.
    0.0054668822923283475

    """
    if t1 > t2:
        raise ValueError("The second threshold should be greater than the first threshold.")
    em = abs(np.linalg.norm(a_local)-g)/g   # Magnitude error (eq. 60)
    f_em = np.clip((t2-em)/t1, 0.0, 1.0)    # Gain factor
    return alpha_bar * f_em   # Filtering gain (eq. 61)

class AQUA:
    """
    Algebraic Quaternion Algorithm

    Parameters
    ----------
    gyr : numpy.ndarray, default: None
        N-by-3 array with measurements of angular velocity in rad/s
    acc : numpy.ndarray, default: None
        N-by-3 array with measurements of acceleration in m/s^2
    mag : numpy.ndarray, default: None
        N-by-3 array with measurements of magnetic field in nT
    frequency : float, default: 100.0
        Sampling frequency in Herz
    Dt : float, default: 0.01
        Sampling step in seconds. Inverse of sampling frequency. Not required
        if ``frequency`` value is given.
    alpha : float, default: 0.01
        Gain characterizing cut-off frequency for accelerometer quaternion
    beta : float, default: 0.01
        Gain characterizing cut-off frequency for magnetometer quaternion
    threshold : float, default: 0.9
        Threshold to discriminate between LERP and SLERP interpolation
    adaptive : bool, default: False
        Whether to use an adaptive gain for each sample
    q0 : numpy.ndarray, default: None
        Initial orientation, as a versor (normalized quaternion).

    Attributes
    ----------
    gyr : numpy.ndarray
        N-by-3 array with N gyroscope samples.
    acc : numpy.ndarray
        N-by-3 array with N accelerometer samples.
    mag : numpy.ndarray
        N-by-3 array with N magnetometer samples.
    frequency : float
        Sampling frequency in Herz
    Dt : float
        Sampling step in seconds. Inverse of sampling frequency.
    alpha : float
        Gain characterizing cut-off frequency for accelerometer quaternion.
    beta : float
        Gain characterizing cut-off frequency for magnetometer quaternion.
    threshold : float
        Threshold to discern between LERP and SLERP interpolation.
    adaptive : bool
        Flag indicating use of adaptive gain.
    q0 : numpy.ndarray
        Initial orientation, as a versor (normalized quaternion).

    """
    def __init__(self, acc: np.ndarray = None, mag: np.ndarray = None, gyr: np.ndarray = None, **kw):
        self.gyr: np.ndarray = gyr
        self.acc: np.ndarray = acc
        self.mag: np.ndarray = mag
        self.frequency: float = kw.get('frequency', 100.0)
        self.frame: str = kw.get('frame', 'NED')
        self.Dt: float = kw.get('Dt', (1.0/self.frequency) if self.frequency else 0.01)
        self.alpha: float = kw.get('alpha', 0.01)
        self.beta: float = kw.get('beta', 0.01)
        self.threshold: float = kw.get('threshold', 0.9)
        self.adaptive: bool = kw.get('adaptive', False)
        self._assert_validity_of_inputs()
        self.q0: np.ndarray = kw.get('q0')
        if self.acc is not None:
            self.Q = self._compute_all()

    def _assert_validity_of_inputs(self):
        """Asserts the validity of the inputs."""
        for item in ["frequency", "Dt", "alpha", "beta", "threshold"]:
            if isinstance(self.__getattribute__(item), bool):
                raise TypeError(f"Parameter '{item}' must be numeric.")
            if not isinstance(self.__getattribute__(item), (int, float)):
                raise TypeError(f"Parameter '{item}' is not a non-zero number.")
            if self.__getattribute__(item) <= 0.0:
                raise ValueError("Parameter '{item}' must be a non-zero number.")
        if not isinstance(self.adaptive, bool):
            raise TypeError(f"Parameter 'adaptive' is not a boolean type. It is {type(self.adaptive)}.")
        if not isinstance(self.frame, str):
            raise TypeError(f"Parameter 'frame' is not a string. It is {type(self.frame)}.")
        if self.frame.upper() not in ['NED', 'ENU']:
            raise ValueError(f"Frame '{self.frame}' is not valid. Try 'NED' or 'ENU'.")

    def _assert_triaxial_sample_vector(self, input_vector: np.ndarray, name: str):
        """Asserts the input vector is a tri-axial sample."""
        _assert_numerical_iterable(input_vector, name)
        input_vector = np.copy(input_vector)
        if input_vector.ndim > 1:
            raise ValueError(f"Sample of {name} must be one tri-axial sample.")
        if input_vector.size != 3:
            raise ValueError(f"Sample of {name} must be a tri-axial sample.")

    def _compute_all(self):
        """Estimate all quaternions with given sensor values."""
        _assert_numerical_iterable(self.acc, 'Accelerometer data')
        # A single sample was given
        if np.array(self.acc).ndim < 2:
            if self.mag is None:
                return self.estimate(self.acc)
            _assert_same_shapes(self.acc, self.mag, ['acc', 'mag'])
            return self.estimate(self.acc, self.mag)
        # Multiple samples were given
        num_samples = len(self.acc)
        Q = np.zeros((num_samples, 4))
        if self.mag is None:
            Q[0] = self.estimate(self.acc[0]) if self.q0 is None else self.q0.copy()
            if self.gyr is not None:
                _assert_numerical_iterable(self.gyr, 'Gyroscope data')
                _assert_same_shapes(self.acc, self.gyr, ['acc', 'gyr'])
                for t in range(1, num_samples):
                    Q[t] = self.updateIMU(Q[t-1], self.gyr[t], self.acc[t])
                return Q
            for t in range(1, num_samples):
                Q[t] = self.estimate(self.acc[t])
            return Q
        Q[0] = self.estimate(self.acc[0], self.mag[0]) if self.q0 is None else self.q0.copy()
        _assert_same_shapes(self.acc, self.mag, ['acc', 'mag'])
        if self.gyr is not None:
            _assert_numerical_iterable(self.mag, 'Magnetometer data')
            _assert_numerical_iterable(self.gyr, 'Gyroscope data')
            _assert_same_shapes(self.acc, self.gyr, ['acc', 'gyr'])
            for t in range(1, num_samples):
                Q[t] = self.updateMARG(Q[t-1], self.gyr[t], self.acc[t], self.mag[t])
            return Q
        for t in range(1, num_samples):
            Q[t] = self.estimate(self.acc[t], self.mag[t])
        return Q

    def Omega(self, x: np.ndarray) -> np.ndarray:
        """
        Omega operator.

        Given a vector :math:`\\mathbf{x}\\in\\mathbb{R}^3`, return a
        :math:`4\\times 4` matrix of the form:

        .. math::
            \\boldsymbol\\Omega(\\mathbf{x}) =
            \\begin{bmatrix}
            0 & -\\mathbf{x}^T \\\\ \\mathbf{x} & \\lfloor\\mathbf{x}\\rfloor_\\times
            \\end{bmatrix} =
            \\begin{bmatrix}
            0 & x_1 & x_2 & x_3 \\\\
            -x_1 & 0 & x_3 & -x_2 \\\\
            -x_2 & -x_3 & 0 & x_1 \\\\
            -x_3 & x_2 & -x_1 & 0
            \\end{bmatrix}

        This operator is a simplification to create a 4-by-4 matrix used for
        the product between the angular rate and a quaternion, such that:

        .. math::

            ^L_G\\dot{\\mathbf{q}}_{\\omega, t_k} = \\frac{1}{2}\\boldsymbol\\Omega(\\,^L\\mathbf{\\omega}_{q, t_k})\\;^L_G\\mathbf{q}_{t_{k-1}}

        .. note::
            The original definition in the article (eq. 39) has an errata
            missing the multiplication with :math:`\\frac{1}{2}`.

        Parameters
        ----------
        x : numpy.ndarray
            Three-dimensional vector representing the angular rate around the
            three axes of the local frame.

        Returns
        -------
        Omega : numpy.ndarray
            Omega matrix.
        """
        return 0.5 * np.array([
            [0.0,    x[0],  x[1],  x[2]],
            [-x[0],   0.0,  x[2], -x[1]],
            [-x[1], -x[2],   0.0,  x[0]],
            [-x[2],  x[1], -x[0],   0.0]])

    def estimate(self, acc: np.ndarray, mag: np.ndarray = None) -> np.ndarray:
        """
        Quaternion from Earth-Field Observations

        Algebraic estimation of a quaternion as a function of an observation of
        the Earth's gravitational force and magnetic fields.

        It decomposes the quaternion :math:`\\mathbf{q}` into two auxiliary
        quaternions :math:`\\mathbf{q}_{\\mathrm{acc}}` and
        :math:`\\mathbf{q}_{\\mathrm{mag}}`, such that:

        .. math::
            \\mathbf{q} = \\mathbf{q}_{\\mathrm{acc}}\\mathbf{q}_{\\mathrm{mag}}

        Parameters
        ----------
        acc : numpy.ndarray
            Sample of tri-axial Accelerometer in m/s^2
        mag : numpy.ndarray, default: None
            Sample of tri-axial Magnetometer in mT

        Returns
        -------
        q : numpy.ndarray
            Estimated quaternion.
        """
        self._assert_triaxial_sample_vector(acc, 'accelerometer')
        a_norm = np.linalg.norm(acc)
        if a_norm == 0:
            raise ValueError("Invalid acceleration data. It must contain at least one non-zero value.")
        ax, ay, az = acc/a_norm
        # Quaternion from Accelerometer Readings (eq. 25)
        if az >= 0:
            q_acc = np.array([np.sqrt((az+1)/2), -ay/np.sqrt(2*(az+1)), ax/np.sqrt(2*(az+1)), 0.0])
        else:
            q_acc = np.array([-ay/np.sqrt(2*(1-az)), np.sqrt((1-az)/2.0), 0.0, ax/np.sqrt(2*(1-az))])
        q_acc = Quaternion(q_acc)
        if mag is not None:
            self._assert_triaxial_sample_vector(mag, 'magnetometer')
            m_norm = np.linalg.norm(mag)
            if m_norm == 0:
                raise ValueError("Invalid geomagnetic field. Its must contain at least one non-zero value.")
            lx, ly, _ = q_acc.to_DCM().T @ (mag/m_norm) # (eq. 26)
            Gamma = lx**2 + ly**2                       # (eq. 28)
            # Quaternion from Magnetometer Readings (eq. 35)
            if lx >= 0:
                q_mag = np.array([np.sqrt(Gamma+lx*np.sqrt(Gamma))/np.sqrt(2*Gamma),
                                  0.0,
                                  0.0,
                                  ly/(np.sqrt(2)*np.sqrt(Gamma+lx*np.sqrt(Gamma)))
                                  ])
            else:
                q_mag = np.array([ly/(np.sqrt(2)*np.sqrt(Gamma-lx*np.sqrt(Gamma))),
                                  0.0,
                                  0.0,
                                  np.sqrt(Gamma-lx*np.sqrt(Gamma))/np.sqrt(2*Gamma)
                                  ])
            # Generalized Quaternion Orientation (eq. 36)
            q = q_acc.product(q_mag)
            return q/np.linalg.norm(q)
        return q_acc

    def updateIMU(self, q: np.ndarray, gyr: np.ndarray, acc: np.ndarray, dt: float = None) -> np.ndarray:
        """
        Quaternion Estimation with IMU architecture.

        The estimation is made in two steps: a *prediction* is done with the
        angular rate (gyroscope) to integrate and estimate the current
        orientation; then a *correction* step uses the measured accelerometer
        to infer the expected gravity vector and use it to correct the
        predicted quaternion.

        If the gyroscope data is invalid, it returns the given a-priori
        quaternion. Secondly, if the accelerometer data is invalid the
        predicted quaternion (using gyroscopes) is returned.

        Parameters
        ----------
        q : numpy.ndarray
            A-priori quaternion.
        gyr : numpy.ndarray
            Sample of tri-axial Gyroscope in rad/s.
        acc : numpy.ndarray
            Sample of tri-axial Accelerometer in m/s^2
        dt : float, default: None
            Time step, in seconds, between consecutive Quaternions.

        Returns
        -------
        q : numpy.ndarray
            Estimated quaternion.

        """
        dt = self.Dt if dt is None else dt
        if gyr is None or np.linalg.norm(gyr) == 0:
            return q
        # PREDICTION
        qDot = self.Omega(gyr) @ q                          # Quaternion derivative (eq. 39)
        qInt = Quaternion(q + qDot*dt)                      # Quaternion integration (eq. 42)
        # CORRECTION
        a_norm = np.linalg.norm(acc)
        if a_norm == 0:
            return qInt.to_array()
        a = acc/a_norm
        gx, gy, gz = qInt.to_DCM().T @ a                    # Predicted gravity (eq. 44)
        q_acc = np.array([np.sqrt((gz+1.0)/2.0), -gy/np.sqrt(2.0*(gz+1.0)), gx/np.sqrt(2.0*(gz+1.0)), 0.0])     # Delta Quaternion (eq. 47)
        if self.adaptive:
            self.alpha = adaptive_gain(acc)
        q_acc = slerp_I(q_acc, self.alpha, self.threshold)
        q_prime = qInt.product(q_acc)                       # (eq. 53)
        return q_prime/np.linalg.norm(q_prime)

    def updateMARG(self, q: np.ndarray, gyr: np.ndarray, acc: np.ndarray, mag: np.ndarray, dt: float = None) -> np.ndarray:
        """
        Quaternion Estimation with MARG architecture.

        The estimation is made in two steps: a *prediction* is done with the
        angular rate (gyroscope) to integrate and estimate the current
        orientation; then a *correction* step uses the measured accelerometer
        and magnetic field to infer the expected geodetic values. Its
        divergence is used to correct the predicted quaternion.

        If the gyroscope data is invalid, it returns the given a-priori
        quaternion. Secondly, if the accelerometer data is invalid the
        predicted quaternion (using gyroscopes) is returned. Finally, if the
        magnetometer measurements are invalid, returns a quaternion corrected
        by the accelerometer only.

        Parameters
        ----------
        q : numpy.ndarray
            A-priori quaternion.
        gyr : numpy.ndarray
            Sample of tri-axial Gyroscope in rad/s.
        acc : numpy.ndarray
            Sample of tri-axial Accelerometer in m/s^2
        mag : numpy.ndarray
            Sample of tri-axial Magnetometer in mT
        dt : float, default: None
            Time step, in seconds, between consecutive Quaternions.

        Returns
        -------
        q : numpy.ndarray
            Estimated quaternion.

        """
        dt = self.Dt if dt is None else dt
        if gyr is None or np.linalg.norm(gyr) == 0:
            return q
        # PREDICTION
        qDot = self.Omega(gyr) @ q                          # Quaternion derivative (eq. 39)
        qInt = Quaternion(q + qDot*dt)                      # Quaternion integration (eq. 42)
        # CORRECTION
        a_norm = np.linalg.norm(acc)
        if a_norm == 0:
            return qInt.to_array()
        a = acc/a_norm
        gx, gy, gz = qInt.to_DCM().T @ a                    # Predicted gravity (eq. 44)
        # Accelerometer-Based Quaternion
        q_acc = np.array([np.sqrt((gz+1.0)/2.0), -gy/np.sqrt(2.0*(gz+1.0)), gx/np.sqrt(2.0*(gz+1.0)), 0.0])     # Delta Quaternion (eq. 47)
        if self.adaptive:
            self.alpha = adaptive_gain(acc)
        q_acc = slerp_I(q_acc, self.alpha, self.threshold)
        q_prime = Quaternion(qInt.product(q_acc))           # (eq. 53)
        # Magnetometer-Based Quaternion
        m_norm = np.linalg.norm(mag)
        if m_norm == 0:
            return q_prime.to_array()
        lx, ly, _ = q_prime.to_DCM().T @ (mag/m_norm)       # World frame magnetic vector (eq. 54)
        Gamma = lx**2 + ly**2                               # (eq. 28)
        q_mag = np.array([np.sqrt(Gamma+lx*np.sqrt(Gamma))/np.sqrt(2*Gamma), 0.0, 0.0, ly/np.sqrt(2*(Gamma+lx*np.sqrt(Gamma)))])    # (eq. 58)
        q_mag = slerp_I(q_mag, self.beta, self.threshold)
        # Generalized Quaternion
        q = q_prime.product(q_mag)                          # (eq. 59)
        return q/np.linalg.norm(q)

    def init_q(self, acc: np.ndarray, mag: np.ndarray = None) -> np.ndarray:
        """Synonym of :meth:`estimate`"""
        return self.estimate(acc, mag)



# -*- coding: utf-8 -*-
"""
Attitude estimation algorithm as proposed in :cite:p:`fourati2011`, whose
approach combines a quaternion-based nonlinear filter with the
`Levenberg-Marquardt Algorithm (LMA.)
<https://en.wikipedia.org/wiki/Levenberg%E2%80%93Marquardt_algorithm>`_.

The estimation algorithm has a complementary structure that exploits
measurements from an accelerometer, a magnetometer and a gyroscope, combined in
a strap-down system, based on the time integral of the angular velocity, using
the Earth's magnetic field and gravity vector to compensate the attitude
predicted by the gyroscope.

The **rigid body attitude** in space is determined when the body's orientation
frame :math:`(X_B, Y_B, Z_B)` is specified with respect to the navigation frame
:math:`(X_N, Y_N, Z_N)`, where the navigation frame follows the NED convention
(North-East-Down.)

The unit quaternion, :math:`\\mathbf{q}`, is defined as a scalar-vector pair of
the form:

.. math::
    \\mathbf{q} = \\begin{pmatrix}s & \\mathbf{v}\\end{pmatrix}^T

where :math:`s` is the scalar part and :math:`\\mathbf{v}=\\begin{pmatrix}v_x & v_y & v_z\\end{pmatrix}^T`
is the vector part of the quaternion.

.. note::
    Most literature, and this Python package's documentation, use the notation
    :math:`\\mathbf{q}=\\begin{pmatrix}q_w & q_x & q_y & q_z\\end{pmatrix}` to
    define a quaternion, but this algorithm uses a different one. The latter
    will preserved in this page to keep coherence with the original document.

The sensor configuration consists of a three-axis gyroscope, a three-axis
accelerometer, a three-axis magnetometer. Their outputs can be modelled,
respectively, as:

.. math::
    \\begin{array}{rcl}
    \\omega_G =& \\begin{bmatrix}\\omega_{GX} & \\omega_{GY} & \\omega_{GZ}\\end{bmatrix}^T &= \\omega + b + \\delta_G \\\\&&\\\\
    \\mathbf{f} =& \\begin{bmatrix}f_x & f_y & f_z\\end{bmatrix}^T &= M_N^B(\\mathbf{q}) (g+a) + \\delta_f \\\\&&\\\\
    \\mathbf{h} =& \\begin{bmatrix}h_x & h_y & h_z\\end{bmatrix}^T &= M_N^B(\\mathbf{q}) m + \\delta_h
    \\end{array}

where :math:`b\\in\\mathbb{R}^3` is the unknown gyro-bias vector and :math:`\\delta_G`,
:math:`\\delta_f` and :math:`\\delta_h\\in\\mathbb{R}^3` are assumed `white
Gaussian noises <https://en.wikipedia.org/wiki/Additive_white_Gaussian_noise>`_.
:math:`\\omega` is the *real* angular velocity, :math:`g` is the gravity vector,
:math:`a` denotes the Dynamic Body Acceleration (DBA), :math:`m` describes the
direction of the Earth's magnetic field on the local position, and
:math:`M_N^B(\\mathbf{q})` is the orthogonal matrix describing the attitude of
the body frame.

.. math::
    \\mathbf{M}_N^B(\\mathbf{q}) =
    \\begin{bmatrix}
    2(s^2 + v_x^2) - 1 & 2(v_xv_y + sv_z) & 2(v_xv_z - sv_y) \\\\
    2(v_xv_y - sv_z) & 2(s^2 + v_y^2) - 1 & 2(sv_x + v_yv_z) \\\\
    2(sv_y + v_xv_z) & 2(v_yv_z - sv_x) & 2(s^2 + v_z^2) - 1
    \\end{bmatrix}

The kinematic differential equation, in terms of the unit quaternion, that
describes the relationship between the rigid body attitude variation and the
angular velocity in the body frame is represented by:

.. math::
    \\begin{array}{rcl}
    \\dot{\\mathbf{q}} &=& \\frac{1}{2}\\mathbf{q}\\omega_\\mathbf{q} \\\\
    \\begin{bmatrix}\\dot{s}\\\\ \\dot{v}_x \\\\ \\dot{v}_y \\\\ \\dot{v}_z\\end{bmatrix}
    &=& \\frac{1}{2}\\begin{bmatrix}-\\mathbf{v}^T \\\\ \\mathbf{I}_3s+\\lfloor\\mathbf{v}\\rfloor_\\times\\end{bmatrix}
    \\begin{bmatrix}\\omega_x \\\\ \\omega_y \\\\ \\omega_z\\end{bmatrix}
    \\end{array}

where :math:`\\omega_\\mathbf{q}=\\begin{bmatrix}0 & \\omega^T\\end{bmatrix}^T`
is the equivalent to the angular velocity :math:`\\omega\\in\\mathbb{R}^3` of
the rigid body measured in :math:`B` and relative to :math:`N`, :math:`\\mathbf{I}_3`
is the :math:`3\\times 3` identity matrix, and :math:`\\lfloor\\mathbf{v}\\rfloor_\\times`
is the `Skew symmetric matrix <https://en.wikipedia.org/wiki/Skew-symmetric_matrix>`_
of the vector :math:`\\mathbf{v}`.

.. math::
    \\lfloor\\mathbf{v}\\rfloor_\\times =
    \\begin{bmatrix}0 & -v_z & v_y \\\\ v_z & 0 & -v_x \\\\ -v_y & v_x & 0\\end{bmatrix}

.. note::
    Any vector :math:`\\mathbf{x}=\\begin{bmatrix}x_1 & x_2 & x_3\\end{bmatrix}^T\\in\\mathbb{R}^3`
    that multiplies with a quaternion must be considered a `pure quaternion
    <https://en.wikipedia.org/wiki/Quaternions_and_spatial_rotation#Using_quaternion_as_rotations>`_,
    :math:`\\mathbf{x_q}=\\begin{bmatrix}0 & x_1 & x_2 & x_3\\end{bmatrix}^T\\in\\mathbb{R}^4`,
    so that they operate with a *Hamilton product*.

To achieve an optimal attitude estimation, a nonlinear system is developed,
whose **output** is the stack of the accelerometer and magnetometer
measurements:

.. math::
    \\mathbf{y} = \\begin{bmatrix}f_x & f_y & f_z & h_x & h_y & h_z\\end{bmatrix}^T

The `World Magnetic Model <../wmm.html>`_ considers a magnetic vector
:math:`\\mathbf{m}=\\begin{bmatrix}m_x & m_y & m_z \\end{bmatrix}\\in\\mathbb{R}^3`
at any location on Earth to describe the geomagnetic field. For practical
purposes, the vector is simplified to
:math:`\\mathbf{m}=\\begin{bmatrix}m\\cos\\theta & 0 & m\\sin\\theta\\end{bmatrix}`,
with a dip angle :math:`\\theta` and a magnetic intensity :math:`m`, which
varies between 23000 and 67000 nT, depending on the region on Earth. This
simplified vector discards the Easterly magnetic field (:math:`m_y`), although
for an accurate reference, it is preferred to use it.

Similar to :math:`\\mathbf{y}`, the estimated values :math:`\\hat{\\mathbf{y}}`
are given by:

.. math::
    \\hat{\\mathbf{y}} = \\begin{bmatrix}\\hat{f}_x & \\hat{f}_y & \\hat{f}_z & \\hat{h}_x & \\hat{h}_y & \\hat{h}_z\\end{bmatrix}^T

whose components are calculated as:

.. math::
    \\begin{array}{rcl}
    \\hat{\\mathbf{f}} &=& \\begin{bmatrix}\\hat{f}_x & \\hat{f}_y & \\hat{f}_z \\end{bmatrix}^T = \\hat{\\mathbf{q}}^{-1}\\mathbf{g_q}\\hat{\\mathbf{q}} \\\\ && \\\\
    \\hat{\\mathbf{h}} &=& \\begin{bmatrix}\\hat{h}_x & \\hat{h}_y & \\hat{h}_z \\end{bmatrix}^T = \\hat{\\mathbf{q}}^{-1}\\mathbf{m_q}\\hat{\\mathbf{q}}
    \\end{array}

where :math:`\\mathbf{g_q}=\\begin{bmatrix}0 & 0 & 0 & 9.8\\end{bmatrix}^T` is
the **reference gravity vector** as a pure quaternion, and
:math:`\\mathbf{m_q}=\\begin{bmatrix}0 & m\\cos\\theta & 0 & m\\sin\\theta\\end{bmatrix}^T`
is the local **reference geomagnetic field** also represented as a pure
quaternion.

The modeling error, :math:`\\delta(\\hat{\\mathbf{q}})=\\mathbf{y}-\\hat{\\mathbf{y}}`,
represents the difference between the real measurements :math:`\\mathbf{y}` and
the estimated values :math:`\\hat{\\mathbf{y}}`.

The nonlinear filter of this model takes the form:

.. math::
    \\dot{\\mathbf{q}} =
    \\begin{bmatrix}\\dot{s}\\\\ \\dot{v}_x \\\\ \\dot{v}_y \\\\ \\dot{v}_z\\end{bmatrix} =
    \\frac{1}{2}\\hat{\\mathbf{q}}\\omega_\\mathbf{q}
    \\begin{bmatrix}1 \\\\ \\mathbf{K}\\end{bmatrix}

where :math:`\\hat{\\mathbf{q}}=\\begin{bmatrix}\\hat{s}& \\hat{v}_x & \\hat{v}_y & \\hat{v}_z\\end{bmatrix}^T\\in\\mathbb{R}^4`
is the **estimated state**, and :math:`\\mathbf{K}\\in\\mathbb{R}^{3\\times 6}`
is the **observer gain**.

This gain :math:`\\mathbf{K}` is used to correct the modeling error
:math:`\\delta(\\hat{\\mathbf{q}})`, which can be done if we locate the minimum
of the squared error function :math:`\\xi(\\hat{\\mathbf{q}})=\\delta(\\hat{\\mathbf{q}})^T\\delta(\\hat{\\mathbf{q}})`.

For this attitude estimator the `Levenberg-Marquardt Algorithm
<https://en.wikipedia.org/wiki/Levenberg%E2%80%93Marquardt_algorithm>`_ (LMA)
is used to minimize the nonlinear function :math:`\\xi(\\hat{\\mathbf{q}})`.
So, the unique minimum can be computed with:

.. math::
    \\begin{array}{rcl}
    \\eta(\\hat{\\mathbf{q}}) &=& \\mathbf{K}\\delta(\\hat{\\mathbf{q}}) \\\\
    &=& k[\\mathbf{X}^T\\mathbf{X} + \\lambda\\mathbf{I}_3]^{-1}\\mathbf{X}^T\\delta(\\hat{\\mathbf{q}})
    \\end{array}

where the tiny value :math:`\\lambda` guarantees the inversion of the matrix,
the gain factor :math:`k` tunes the balance between measurement noise
supression and the filter's response time, and
:math:`\\mathbf{X}\\in\\mathbb{R}^{6\\times 3}` is the Jacobian matrix:

.. math::
    \\begin{array}{rcl}
    \\mathbf{X} &=& -2\\begin{bmatrix}\\lfloor\\hat{\\mathbf{f}}\\rfloor_\\times & \\lfloor\\hat{\\mathbf{h}}\\rfloor_\\times\\end{bmatrix}^T \\\\
    &=& -2\\begin{bmatrix}
    0 & -\\hat{f}_z & \\hat{f}_y & 0 & -\\hat{h}_z & \\hat{h}_y \\\\
    \\hat{f}_z & 0 & -\\hat{f}_x & \\hat{h}_z & 0 & -\\hat{h}_x \\\\
    -\\hat{f}_y & \\hat{f}_x & 0 & -\\hat{h}_y & \\hat{h}_x & 0
    \\end{bmatrix}^T
    \\end{array}

The resulting structure of the nonlinear filter is complementary: it blends the
low-frequency region (low bandwidth) of the accelerometer and magnetometer data,
where the attitude is typically more accurate, with the high-frequency region
(high bandwidth) of the gyroscope data, where the integration of the angular
velocity yields better attitude estimates.

By filtering the high-frequency components of the signals from the
accelerometer (DBA) and the low-frequency components of the gyroscope signal
(slow-moving drift), the nonlinear filter produces an accurate estimate of the
attitude.

The correction term, :math:`\\Delta\\in\\mathbb{R}^{4\\times 7}`, is computed
using the gain :math:`K` such as:

.. math::
    \\Delta =
    \\begin{bmatrix}1 & \\mathbf{0} \\\\ \\mathbf{0} & \\mathbf{K}\\end{bmatrix}
    \\begin{bmatrix}1 \\\\ \\delta(\\hat{\\mathbf{q}})\\end{bmatrix}

It is used to correct the estimated angular velocity, :math:`\\dot{\\hat{\\mathbf{q}}}`,
as:

.. math::
    \\dot{\\hat{\\mathbf{q}}} = \\big(\\frac{1}{2}\\hat{\\mathbf{q}}\\omega_\\mathbf{q}\\big)\\Delta

With the corrected angular velocity, we integrate it using the sampling step
:math:`\\Delta_t` and add it to the previous quaternion :math:`\\mathbf{q}_{t-1}`
to obtain the new attitude :math:`\\mathbf{q}_t`:

.. math::
    \\mathbf{q}_t = \\mathbf{q}_{t-1} + \\dot{\\hat{\\mathbf{q}}}\\Delta_t

.. warning::
    Do not confuse the correction term :math:`\\Delta` with the sampling step
    :math:`\\Delta_t`, which is actually the inverse of the sampling frequency
    :math:`f=\\frac{1}{\\Delta_t}`.

"""

from typing import Union
import numpy as np
from ..common.quaternion import Quaternion
from ..common.orientation import ecompass
from ..common.constants import MUNICH_LONGITUDE
from ..common.constants import MUNICH_LATITUDE
from ..common.constants import MUNICH_HEIGHT
from ..common.mathfuncs import cosd
from ..common.mathfuncs import sind
from ..common.mathfuncs import skew
from ..utils.core import _assert_numerical_iterable
from ..utils.core import _assert_numerical_positive_variable

# Reference Observations in Munich, Germany
from ..utils.wmm import WMM
wmm = WMM(latitude=MUNICH_LATITUDE, longitude=MUNICH_LONGITUDE, height=MUNICH_HEIGHT)
REFERENCE_MAGNETIC_VECTOR = np.array([wmm.X, wmm.Y, wmm.Z])

def _set_magnetic_field_vector(magnetic_dip: Union[int, float, list, tuple, np.ndarray]):
    """
    Set the magnetic reference vector.

    Parameters
    ----------
    magnetic_dip : float, array-like
        Magnetic dip, in degrees, or local geomagnetic reference as
        three-dimensional vector.

    """
    if isinstance(magnetic_dip, bool):
        raise TypeError("magnetic_dip must be given as a float, list, tuple or NumPy array. Got bool")
    elif isinstance(magnetic_dip, (float, int)):
        magnetic_field = np.array([0.0, cosd(magnetic_dip), 0., sind(magnetic_dip)])
    elif isinstance(magnetic_dip, (list, tuple, np.ndarray)):
        if not all(isinstance(x, (float, int)) for x in magnetic_dip):
            raise TypeError("magnetic_dip must be an array of floats. Contains non-numeric values.")
        magnetic_field = np.copy(magnetic_dip)
    elif magnetic_dip is None:
        magnetic_field = np.array([0.0, *REFERENCE_MAGNETIC_VECTOR])
    else:
        raise TypeError(f"magnetic_dip must be given as a float, list, tuple or NumPy array. Got {type(magnetic_dip)}")
    if magnetic_field.shape != (4,):
        raise ValueError(f"magnetic_dip array must contain 4 elements. Got {magnetic_field.shape}")
    return magnetic_field / np.linalg.norm(magnetic_field)

class Fourati:
    """
    Fourati's attitude estimation

    Parameters
    ----------
    acc : numpy.ndarray, default: None
        N-by-3 array with measurements of acceleration in in m/s^2
    gyr : numpy.ndarray, default: None
        N-by-3 array with measurements of angular velocity in rad/s
    mag : numpy.ndarray, default: None
        N-by-3 array with measurements of magnetic field in mT
    frequency : float, default: 100.0
        Sampling frequency in Herz.
    Dt : float, default: 0.01
        Sampling step in seconds. Inverse of sampling frequency. Not required
        if `frequency` value is given.
    gain : float, default: 0.1
        Filter gain factor.
    q0 : numpy.ndarray, default: None
        Initial orientation, as a versor (normalized quaternion).
    magnetic_dip : float, array-like, default: None
        Magnetic Inclination angle, in degrees, or as a local geomagnetic
        vector in a pure quaternion of the form [0, mx, my, mz], where mx, my,
        and mz are the three-dimensional components of the local geomagnetic
        vector.

    Raises
    ------
    ValueError
        When dimension of input array(s) ``acc``, ``gyr``, or ``mag`` are not equal.

    Examples
    --------
    >>> gyro_data.shape, acc_data.shape, mag_data.shape      # NumPy arrays with sensor data
    ((1000, 3), (1000, 3), (1000, 3))
    >>> from ahrs.filters import Fourati
    >>> fourati = Fourati(gyr=gyro_data, acc=acc_data, mag=mag_data)
    >>> fourati.Q       # Estimated attitudes as Quaternions
    array([[-0.82311077,  0.45760535, -0.33408929, -0.0383452 ],
           [-0.82522048,  0.4547043 , -0.33277675, -0.03892033],
           [-0.82463698,  0.4546915 , -0.33422422, -0.03903417],
           ...,
           [-0.82420642,  0.56217735,  0.02548005, -0.06317571],
           [-0.82364606,  0.56311099,  0.0241655 , -0.06268338],
           [-0.81844766,  0.57077781,  0.02532182, -0.06095017]])
    >>> fourati.Q.shape
    (1000, 4)

    """
    def __init__(self, gyr: np.ndarray = None, acc: np.ndarray = None, mag: np.ndarray = None, **kwargs):
        self.gyr: np.ndarray = gyr
        self.acc: np.ndarray = acc
        self.mag: np.ndarray = mag
        self.frequency: float = kwargs.get('frequency', 100.0)
        self.Dt: float = kwargs.get('Dt', 1.0/self.frequency if self.frequency else 0.01)
        self.gain: float = kwargs.get('gain', 0.1)
        self.q0: np.ndarray = kwargs.get('q0')
        # Reference vectors
        self.m_q: Quaternion = Quaternion(_set_magnetic_field_vector(kwargs.get('magnetic_dip')))   # Gravity vector as pure quaternion
        self.g_q: Quaternion = Quaternion([0.0, 0.0, 0.0, 1.0])                                     # Magnetic field vector as pure quaternion
        # Process of given data
        self._assert_validity_of_inputs()
        if self.acc is not None and self.gyr is not None and self.mag is not None:
            self.Q = self._compute_all()

    def _assert_validity_of_inputs(self):
        """Asserts the validity of the inputs."""
        for item in ["frequency", "Dt", "gain"]:
            _assert_numerical_positive_variable(getattr(self, item), item)
        if self.q0 is not None:
            self.q0 = Quaternion(self.q0).to_array()

    def _compute_all(self):
        """
        Estimate the quaternions given all data

        Attributes ``gyr``, ``acc`` and ``mag`` must contain data.

        Returns
        -------
        Q : numpy.ndarray
            M-by-4 Array with all estimated quaternions, where M is the number
            of samples.

        """
        _assert_numerical_iterable(self.gyr, 'Angular velocity vector')
        _assert_numerical_iterable(self.acc, 'Gravitational acceleration vector')
        _assert_numerical_iterable(self.mag, 'Geomagnetic field vector')
        self.gyr = np.copy(self.gyr)
        self.acc = np.copy(self.acc)
        self.mag = np.copy(self.mag)
        if self.acc.shape != self.gyr.shape:
            raise ValueError("acc and gyr are not the same size")
        if self.mag.shape != self.gyr.shape:
            raise ValueError("mag and gyr are not the same size")
        num_samples = len(self.gyr)
        Q = np.zeros((num_samples, 4))
        Q[0] = ecompass(self.acc[0], self.mag[0], frame='NED', representation='quaternion')
        for t in range(1, num_samples):
            Q[t] = self.update(Q[t-1], self.gyr[t], self.acc[t], self.mag[t])
        return Q

    def update(self, q: np.ndarray, gyr: np.ndarray, acc: np.ndarray, mag: np.ndarray, dt: float = None) -> np.ndarray:
        """
        Quaternion Estimation with a MARG architecture.

        Parameters
        ----------
        q : numpy.ndarray
            A-priori quaternion.
        gyr : numpy.ndarray
            Sample of tri-axial Gyroscope in rad/s
        acc : numpy.ndarray
            Sample of tri-axial Accelerometer in m/s^2
        mag : numpy.ndarray
            Sample of tri-axial Magnetometer in mT
        dt : float, default: None
            Time step, in seconds, between consecutive Quaternions.

        Returns
        -------
        q : numpy.ndarray
            Estimated quaternion.

        """
        _assert_numerical_iterable(q, 'Quaternion')
        _assert_numerical_iterable(gyr, 'Tri-axial gyroscope sample')
        _assert_numerical_iterable(acc, 'Tri-axial accelerometer sample')
        _assert_numerical_iterable(mag, 'Tri-axial magnetometer sample')
        dt = self.Dt if dt is None else dt
        if gyr is None or np.linalg.norm(gyr) == 0:
            return q
        q = Quaternion(q)
        qDot = 0.5 * q.product([0, *gyr])                           # (eq. 5)
        if self.gain > 0:
            a_norm = np.linalg.norm(acc)
            if a_norm == 0:
                raise ValueError("Accelerometer data is null. Cannot estimate quaternion.")
            m_norm = np.linalg.norm(mag)
            if m_norm == 0:
                raise ValueError("Magnetometer data is null. Cannot estimate quaternion.")
            # Levenberg Marquardt Algorithm
            fhat = Quaternion(q.conj).product(self.g_q.product(q))  # (eq. 21)
            hhat = Quaternion(q.conj).product(self.m_q.product(q))  # (eq. 22)
            y = np.r_[acc/a_norm, mag/m_norm]                       # Measurements (eq. 6)
            yhat = np.r_[fhat[1:], hhat[1:]]                        # Estimated values (eq. 8)
            dq = y - yhat                                           # Modeling Error
            X = -2*np.c_[skew(fhat[1:]), skew(hhat[1:])].T          # Jacobian Matrix (eq. 23)
            lam = 1e-5                                              # Deviation to guarantee inversion
            K = self.gain*np.linalg.inv(X.T@X + lam*np.eye(3))@X.T  # Gain (eq. 24)
            eta = K @ dq                                            # Unique minimum (eq. 10)
            Delta = [1, *eta]                                       # Correction term (eq. 25)
            qDot = Quaternion(qDot, versor=False).product(Delta)    # Corrected quaternion rate (eq. 7)
        q += qDot*dt
        return q/np.linalg.norm(q)



# -*- coding: utf-8 -*-
"""
AHRS
====

The versioning follows the principles of Semantic Versioning as specified in
https://semver.org

"""

from . import common
from . import filters
from . import utils
from .common.constants import *
from .common.quaternion import Quaternion
from .common.quaternion import QuaternionArray
from .common.dcm import DCM
from .utils.sensors import Sensors

__version__ = "0.4.0"



# Configuration file for the Sphinx documentation builder.

# -- Path setup --------------------------------------------------------------
import os
import sys
from datetime import datetime
sys.path.insert(0, os.path.abspath('../../ahrs' if sys.platform.startswith('win') else '../..'))
from ahrs import __version__

# -- Project information -----------------------------------------------------
project = 'AHRS'
author = 'Mario Garcia'
copyright = f"2019-{datetime.now().year}, {author}"
release = __version__

# -- General configuration ---------------------------------------------------
extensions = [
    'sphinxcontrib.bibtex',
    'sphinx.ext.autodoc',
    'sphinx.ext.mathjax',
    'sphinx.ext.napoleon',
]

# Napoleon settings
napoleon_numpy_docstring = True
napoleon_use_ivar = True            # List attributes with :ivar:
napoleon_use_param = False

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# The suffix of source filenames.
source_suffix = '.rst'

# Explicitly assign master document
master_doc = 'index'

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
exclude_patterns = ['build']

# -- Options for HTML output -------------------------------------------------
html_theme = "pydata_sphinx_theme"
html_static_path = ['_static']
html_logo = "ahrs_logo.png"
html_favicon = "ahrs_icon.ico"
html_css_files = []  # Explicitly set to an empty list

# Bibliography file
bibtex_bibfiles = ['refs.bib']



#!/usr/bin/env python3
import unittest
import numpy as np
import ahrs

GENERATOR = np.random.default_rng(42)
THRESHOLD = 1e-6
SQRT2_2 = np.sqrt(2)/2

class TestDCM(unittest.TestCase):
    def setUp(self) -> None:
        self.R0 = ahrs.DCM()
        self.vector = GENERATOR.random(3)*180-90
        self.R = ahrs.DCM(rpy=self.vector)
        self.q = ahrs.Quaternion(GENERATOR.random(4)-0.5)
        self.R1 = ahrs.DCM(q=self.q)
        self.multi_R = ahrs.DCM(np.atleast_2d([self.R0, self.R, self.R1]))

    def test_rotation_matrix_in_SO3(self):
        self.assertAlmostEqual(np.linalg.det(self.R), 1.0)
        np.testing.assert_almost_equal(self.R@self.R.T, np.identity(3))
        self.assertRaises(ValueError, ahrs.DCM, -self.R)
        self.assertRaises(ValueError, ahrs.DCM, self.R[:2, :2])
        # Test construction of array with many rotation matrices
        self.assertTrue(all(np.allclose(self.multi_R[i].T, rt) for i, rt, in enumerate(np.transpose(self.multi_R, (0, 2, 1)))))
        self.assertRaises(ValueError, ahrs.DCM, GENERATOR.random((4, 3, 3)) - 0.5)
        self.assertRaises(ValueError, ahrs.DCM, GENERATOR.random((4, 2, 2)) - 0.5)
        self.assertRaises(ValueError, ahrs.DCM, np.zeros((4, 3, 3)))
        self.assertRaises(ValueError, ahrs.DCM, np.ones((4, 3, 3)))

    def test_rotation_matrix_from_euler_angles(self):
        np.testing.assert_almost_equal(ahrs.DCM(x=0.0), np.identity(3))
        np.testing.assert_almost_equal(ahrs.DCM(y=0.0), np.identity(3))
        np.testing.assert_almost_equal(ahrs.DCM(z=0.0), np.identity(3))
        rmx45 = np.array([[1.0, 0.0, 0.0], [0.0, SQRT2_2, -SQRT2_2], [0.0, SQRT2_2, SQRT2_2]])
        rmy45 = np.array([[SQRT2_2, 0.0, SQRT2_2], [0.0, 1.0, 0.0], [-SQRT2_2, 0.0, SQRT2_2]])
        rmz45 = np.array([[SQRT2_2, -SQRT2_2, 0.0], [SQRT2_2, SQRT2_2, 0.0], [0.0, 0.0, 1.0]])
        np.testing.assert_almost_equal(ahrs.DCM(x=45*ahrs.DEG2RAD), rmx45)
        np.testing.assert_almost_equal(ahrs.DCM(y=45*ahrs.DEG2RAD), rmy45)
        np.testing.assert_almost_equal(ahrs.DCM(z=45*ahrs.DEG2RAD), rmz45)
        np.testing.assert_almost_equal(ahrs.DCM(euler=('x', [45*ahrs.DEG2RAD])), rmx45)
        np.testing.assert_almost_equal(ahrs.DCM(euler=('y', [45*ahrs.DEG2RAD])), rmy45)
        np.testing.assert_almost_equal(ahrs.DCM(euler=('z', [45*ahrs.DEG2RAD])), rmz45)
        np.testing.assert_almost_equal(ahrs.DCM(euler=(None, [45*ahrs.DEG2RAD])), rmz45)
        np.testing.assert_almost_equal(ahrs.DCM(euler=(0, [45*ahrs.DEG2RAD])), rmx45)
        np.testing.assert_almost_equal(ahrs.DCM(euler=(1, [45*ahrs.DEG2RAD])), rmy45)
        np.testing.assert_almost_equal(ahrs.DCM(euler=(2, [45*ahrs.DEG2RAD])), rmz45)

    def test_identity_rotation_matrix(self):
        np.testing.assert_equal(self.R0, np.identity(3))
        np.testing.assert_equal(ahrs.DCM(), np.identity(3))
        np.testing.assert_equal(ahrs.DCM(x=0.0), np.identity(3))
        np.testing.assert_equal(ahrs.DCM(y=0.0), np.identity(3))
        np.testing.assert_equal(ahrs.DCM(z=0.0), np.identity(3))
        np.testing.assert_almost_equal(ahrs.DCM(x=360.0*ahrs.DEG2RAD), np.identity(3))
        np.testing.assert_almost_equal(ahrs.DCM(y=360.0*ahrs.DEG2RAD), np.identity(3))
        np.testing.assert_almost_equal(ahrs.DCM(z=360.0*ahrs.DEG2RAD), np.identity(3))

    def test_wrong_input_matrix(self):
        self.assertRaises(TypeError, ahrs.DCM, 3)
        self.assertRaises(TypeError, ahrs.DCM, 3.0)
        self.assertRaises(TypeError, ahrs.DCM, "np.eye(3)")
        self.assertRaises(ValueError, ahrs.DCM, np.random.random((3, 3)))
        self.assertRaises(ValueError, ahrs.DCM, -np.identity(3))
        self.assertRaises(ValueError, ahrs.DCM, np.identity(4))

    def test_wrong_rpy(self):
        self.assertRaises(TypeError, ahrs.DCM, rpy=45.0)
        self.assertRaises(TypeError, ahrs.DCM, rpy=45)
        self.assertRaises(TypeError, ahrs.DCM, rpy=True)
        self.assertRaises(TypeError, ahrs.DCM, rpy="25.0")
        self.assertRaises(TypeError, ahrs.DCM, rpy=["10.0", "20.0", "30.0"])
        self.assertRaises(ValueError, ahrs.DCM, rpy=np.random.random(4))

    def test_wrong_euler(self):
        self.assertRaises(TypeError, ahrs.DCM, euler=1)
        self.assertRaises(TypeError, ahrs.DCM, euler=1.0)
        self.assertRaises(TypeError, ahrs.DCM, euler=True)
        self.assertRaises(TypeError, ahrs.DCM, euler="x")
        self.assertRaises(TypeError, ahrs.DCM, euler=["x", "y", "z"])
        self.assertRaises(TypeError, ahrs.DCM, euler=np.random.random(4))
        self.assertRaises(TypeError, ahrs.DCM, euler=([], 3))
        self.assertRaises(ValueError, ahrs.DCM, euler=(3,))

    def test_wrong_quaternion(self):
        self.assertRaises(TypeError, ahrs.DCM, q=3)
        self.assertRaises(TypeError, ahrs.DCM, q=3.0)
        self.assertRaises(TypeError, ahrs.DCM, q="np.eye(3)")
        self.assertRaises(ValueError, ahrs.DCM, q=np.random.random(3))
        self.assertRaises(ValueError, ahrs.DCM, q=np.random.random(5))
        self.assertRaises(ValueError, ahrs.DCM, q=np.random.random((5, 3)))

    def test_dcm_to_q_default(self):
        self.assertLessEqual(ahrs.utils.metrics.qad(self.q, self.R1.to_quaternion()), THRESHOLD)

    def test_dcm_to_q_chiavereini(self):
        self.assertLessEqual(ahrs.utils.metrics.qad(self.q, self.R1.to_quaternion(method='chiaverini')), THRESHOLD)

    def test_dcm_to_q_hughes(self):
        self.assertLessEqual(ahrs.utils.metrics.qad(self.q, self.R1.to_quaternion(method='hughes')), THRESHOLD)

    def test_dcm_to_q_itzhack(self):
        self.assertLessEqual(ahrs.utils.metrics.qad(self.q, self.R1.to_quaternion(method='itzhack')), THRESHOLD)

    def test_dcm_to_q_itzhack_v1(self):
        self.assertLessEqual(ahrs.utils.metrics.qad(self.q, self.R1.to_quaternion(method='itzhack', version=1)), THRESHOLD)

    def test_dcm_to_q_itzhack_v2(self):
        self.assertLessEqual(ahrs.utils.metrics.qad(self.q, self.R1.to_quaternion(method='itzhack', version=2)), THRESHOLD)

    def test_dcm_to_q_itzhack_v3(self):
        self.assertLessEqual(ahrs.utils.metrics.qad(self.q, self.R1.to_quaternion(method='itzhack', version=3)), THRESHOLD)

    def test_dcm_to_q_sarabandi(self):
        self.assertLessEqual(ahrs.utils.metrics.qad(self.q, self.R1.to_quaternion(method='sarabandi')), THRESHOLD)

    def test_dcm_to_q_shepperd(self):
        self.assertLessEqual(ahrs.utils.metrics.qad(self.q, self.R1.to_quaternion(method='shepperd')), THRESHOLD)

if __name__ == "__main__":
    unittest.main()



# -*- coding: utf-8 -*-
"""
Common mathematical routines
============================

These functions can be used at different scripts and submodules.

"""

import numpy as np
from .constants import DEG2RAD

def cosd(x):
    """
    Return the cosine of `x`, which is expressed in degrees.

    If `x` is a list, it will be converted first to a NumPy array, and then the
    cosine operation over each value will be carried out.

    Parameters
    ----------
    x : float or array-like
        Angle in Degrees.

    Returns
    -------
    y : float or numpy.ndarray
        Cosine of given angle.

    Examples
    --------
    >>> from ahrs.common.mathfuncs import cosd
    >>> cosd(0.0)
    1.0
    >>> cosd(90.0)
    0.0
    >>> cosd(-120.0)
    -0.5

    """
    if isinstance(x, (list, tuple, np.ndarray)):
        x = np.copy(x)
    return np.cos(x*DEG2RAD)

def sind(x):
    """
    Return the sine of `x`, which is expressed in degrees.

    If `x` is a list, it will be converted first to a NumPy array, and then the
    sine operation over each value will be carried out.

    Parameters
    ----------
    x : float or array-like
        Angle in Degrees.

    Returns
    -------
    y : float or numpy.ndarray
        Sine of given angle.

    Examples
    --------
    >>> from ahrs.common.mathfuncs import sind
    >>> sind(0.0)
    0.0
    >>> sind(90.0)
    1.0
    >>> sind(-120.0)
    -0.86602540378

    """
    if isinstance(x, (list, tuple, np.ndarray)):
        x = np.copy(x)
    return np.sin(x*DEG2RAD)

def skew(x):
    """
    Return the 3-by-3 skew-symmetric matrix :cite:p:`Wiki_skew` of a 3-element
    vector ``x``.

    Parameters
    ----------
    x : array-like
        3-element array with values to be ordered in a skew-symmetric matrix.

    Returns
    -------
    X : numpy.ndarray
        3-by-3 numpy array of the skew-symmetric matrix.

    Examples
    --------
    >>> from ahrs.common.mathfuncs import skew
    >>> a = [1, 2, 3]
    >>> skew(a)
    [[ 0. -3.  2.]
     [ 3.  0. -1.]
     [-2.  1.  0.]]
    >>> a = np.array([[4.0], [5.0], [6.0]])
    >>> skew(a)
    [[ 0. -6.  5.]
     [ 6.  0. -4.]
     [-5.  4.  0.]]

    """
    if isinstance(x, (list, tuple, np.ndarray)):
        x = np.copy(x)
    if len(x) != 3:
        raise ValueError("Input must be an array with three elements")
    return np.array([[0, -x[2], x[1]], [x[2], 0, -x[0]], [-x[1], x[0], 0.0]])



# -*- coding: utf-8 -*-
"""
Tools for the Juypter notebooks used to showcase the functionality of the
package.

Many funcitons are based on the ones included in the Python package `ezview`
available in the repository https://https://github.com/Mayitzin/ezview
"""

import copy
from inspect import getmembers, isfunction
import numpy as np
import matplotlib.pyplot as plt

plt.rcParams['grid.color'] = (0.5, 0.5, 0.5, 0.1)  # Light grey grid

def hex_to_int(color: str) -> tuple:
    """Convert a hex color to a tuple of integers."""
    a = color.lstrip('#')
    return tuple(int(a[i:i+2], 16) for i in (0, 2, 4, 6))

def hex_to_float(color: str) -> tuple:
    """Convert a hex color to a tuple of floats."""
    a = color.lstrip('#')
    return tuple(int(a[i:i+2], 16)/255.0 for i in (0, 2, 4, 6))

COLORS = [
    "#FF0000FF", "#00AA00FF", "#0000FFFF", "#999933FF",
    "#FF8888FF", "#88AA88FF", "#8888FFFF", "#999955FF",
    "#660000FF", "#005500FF", "#000088FF", "#666600FF"]
COLORS_INTS = [hex_to_int(c) for c in COLORS]
COLORS_FLOATS = [hex_to_float(c) for c in COLORS]

def describe_methods(obj):
    """Return the description of the methods of an object."""
    for method_name in [x for x in dir(obj) if hasattr(obj.__getattribute__(x), '__call__') and not x.startswith('__')]:
        class_and_method = f"{obj.__class__.__name__}.{method_name+'()': <20}"
        method_description = '<No description found>'
        try:
            method_docstring = getattr(obj, method_name).__doc__
            if isinstance(method_docstring, str):
                method_description = method_docstring.split('\n')[1].strip()
                if len(method_description) < 1:
                    method_description = [x.strip() for x in method_docstring.split('\n') if len(x) > 0][1]
            print(class_and_method + f"{method_description.replace(':meth:', '')}")
        except Exception as e:
            print(class_and_method + f"{method_description} ({e})")

def describe_functions(module):
    """Return the description of the functions of a module."""
    for function_name in [item[0] for item in getmembers(module, isfunction) if not item[0].startswith('_')]:
        function_description = '<No description found>'
        try:
            function_docstring = getattr(module, function_name).__doc__
            if isinstance(function_docstring, str):
                function_description = function_docstring.split('\n')[1].strip()
                if len(function_description) < 1:
                    function_description = [x.strip() for x in function_docstring.split('\n') if len(x) > 0][1]
            print(f"{module.__name__}.{function_name+'()': <24}" + f"{function_description.replace(':meth:', '')}")
        except Exception as e:
            print(f"{module.__name__}.{function_name+'()': <24}" + f"{function_description} ({e})")

def plot(*data, **kw):
    """
    Plot time-series data.

    Parameters
    ----------
    data : array
        Arrays with the contents of data to plot. They could be 1- (single line)
        or 2-dimensional.
    title : int or str
        Window title as number or label.
    subtitles : list
        List of strings of the titles of each subplot.
    labels : list
        List of labels that will be displayed in each subplot's legend.
    xlabels : list
        List of strings of the labels of each subplot's X-axis.
    ylabels : list
        List of strings of the labels of each subplot's Y-axis.
    yscales : str
        List of strings of the scales of each subplot's Y-axis. It supports
        matlabs defaults values: "linear", "log", "symlog" and "logit"

    """
    title = kw.get("title")
    subtitles = kw.get("subtitles")
    labels = kw.get("labels")
    xlabels = kw.get("xlabels")
    ylabels = kw.get("ylabels")
    yscales = kw.get("yscales")
    index = kw.get("index")
    indices = kw.get("indices")
    shades_spans = kw.get("shaded")
    num_subplots = len(data)        # Number of given arrays
    # Create figure with vertically stacked subplots
    fig, axs = plt.subplots(
        num_subplots,
        1,
        num=title,
        squeeze=False,
        sharex=kw.get('sharex', "indices" not in kw),
        sharey=kw.get('sharey', False)
        )
    for i, array in enumerate(data):
        array = np.copy(array)
        if array.ndim > 2:
            raise ValueError(f"Data array {i} has more than 2 dimensions.")
        if array.ndim < 2:
            # Plot a single line in the subplot (1-dimensional array)
            label = labels[i][0] if labels else None
            index = index if index is not None else np.arange(array.shape[0])
            axs[i, 0].plot(index, array, color=COLORS[0], lw=0.5, ls='-', label=label)
        else:
            # Plot multiple lines in the subplot (2-dimensional array)
            array_sz = array.shape
            if array_sz[0] > array_sz[1]:
                # Transpose array if it has more rows than columns
                array = array.T
            index = indices[i] if indices is not None else np.arange(array_sz[0])
            for j, row in enumerate(array):
                label = None
                if labels:
                    if len(labels[i]) == len(array):
                        label = labels[i][j]
                axs[i, 0].plot(index, row, color=COLORS[j], lw=0.5, ls='-', label=label)
        axs[i, 0].grid(axis='y')
        if subtitles:
            axs[i, 0].set_title(subtitles[i])
        if xlabels:
            axs[i, 0].set_xlabel(xlabels[i])
        if ylabels:
            axs[i, 0].set_ylabel(ylabels[i])
        if yscales:
            axs[i, 0].set_yscale(yscales[i])
        if shades_spans is not None:
            # Add shaded areas
            try:
                if isinstance(shades_spans, (list, np.ndarray)):
                    current_spans = shades_spans[i] if np.copy(shades_spans).ndim > 2 else shades_spans
                    for s in current_spans:
                        axs[i, 0].axvspan(s[0], s[1], color='gray', alpha=0.1)
                elif isinstance(shades_spans, dict):
                    # Add shades AND their corresponding labels
                    for k, v in shades_spans.items():
                        span = [v['start'], v['stop']]
                        axs[i, 0].axvspan(span[0], span[1], color='gray', alpha=0.1)
                        axs[i, 0].text(int(np.mean(span)), max(array), k, ha='center')
            except:
                print("No spans were given")
        if labels:
            if len(labels[i]) > 0:
                axs[i, 0].legend(loc='lower right')
    fig.tight_layout()
    plt.show()

def ellipsoid(center: np.ndarray = None, axes: np.ndarray = None, num_points: int = 20) -> tuple:
    """
    Return the mesh of an ellipsoid.

    Parameters
    ----------
    center : numpy.ndarray, optional
        3-element array with the ellipsoid's center. Default is [0, 0, 0].
    axes : numpy.ndarray, optional
        3-element array with the ellipsoid's main axes lengths. Default is
        [1, 1, 1].
    num_points : int, optional
        Number of points to use in the mesh. Default is 20.

    Returns
    -------
    tuple
        Tuple with the mesh of the ellipsoid in the form (x, y, z).

    """
    if center is None:
        center = np.zeros(3)
    if axes is None:
        axes = np.ones(3)
    if not isinstance(center, (np.ndarray, list, tuple)):
        raise TypeError("Center must be a 3-element array.")
    if not isinstance(axes, (np.ndarray, list, tuple)):
        raise TypeError("Axes must be a 3-element array.")
    # Create ellipsoid mesh
    cx, cy, cz = center
    sx, sy, sz = axes
    u, v = np.mgrid[0:2*np.pi:complex(num_points), 0:np.pi:complex(num_points)]
    x = sx * np.cos(u)*np.sin(v) + cx
    y = sy * np.sin(u)*np.sin(v) + cy
    z = sz * np.cos(v) + cz
    return x, y, z

def frame(dcm: np.ndarray = None, position: np.ndarray = None, scale: float = 1.0) -> list:
    """
    Return the coordinates of an orthogonal frame.

    Parameters
    ----------
    dcm : numpy.ndarray, optional
        3-by-3 array with the frame's axes. Default is the identity matrix.
    position : numpy.ndarray, optional
        3-element array with the frame's origin. Default is [0, 0, 0].
    scale : float, optional
        Scale factor for the frame's axes. Default is 1.0.

    """
    if dcm is None:
        dcm = np.identity(3)
    if position is None:
        position = np.zeros(3)
    f_coords = []
    for column_index in range(3):
        axis_end = dcm[:, column_index]*scale + position
        f_coords.append(np.c_[position, axis_end])
    return f_coords


def add_ellipsoid(ax, params: list | dict, num_points: int = 20, color = 'k', lw = 0.5, **kwargs) -> None:
    """
    Add a ellipsoid to an existing 3D plot.

    Parameters
    ----------
    ax : matplotlib.axes.Axes3D
        3D axis where the ellipsoid will be added.
    params : list or dict.
        List or dictionary with the parameters to draw an ellipsoid. If a list
        is given, it must be of the form [[a, b, c], [x, y, z]], where a, b, c
        are the coordinates of the ellipsoid's center, and x, y, z are the
        ellipsoid's main axes lengths. If a dictionary is given, it must be of
        the form {'center': [a, b, c], 'axes': [x, y, z]}
    num_points : int, optional
        Number of points, per axis, to use in the mesh. Default is 20.
    color : str, optional
        Color of the ellipsoid. Default is 'k'.
    lw : float, optional
        Line width of the ellipsoid. Default is 0.5.

    """
    if isinstance(params, (list, tuple, np.ndarray)):
        center, axes = params
    elif isinstance(params, dict):
        center = params.get("center", np.zeros(3))
        axes = params.get("axes", np.ones(3))
    else:
        raise TypeError("Unknown type for 'sphere'. Try a list or a dict.")
    # Extract only the expected parameters from kwargs
    expected_params = {'num_points': num_points, 'color': color, 'lw': lw, 'alpha': 0.5}
    for key in expected_params:
        if key in kwargs:
            expected_params[key] = kwargs[key]
    x, y, z = ellipsoid(center=center, axes=axes, num_points=expected_params['num_points'])   # Ellipsoid mesh
    ax.plot_wireframe(x, y, z, color=expected_params['color'], lw=expected_params['lw'], alpha=expected_params['alpha'])

def add_frame(ax, dcm: np.ndarray, position = None, color: str | list = None, scale: float = 1.0, lw: float = 1.5, **kwargs) -> None:
    """
    Add a frame to an existing 3D plot.

    Parameters
    ----------
    ax : mpl_toolkits.mplot3d.axes3d.Axes3D
        3D axis where the frame will be added.
    frame : numpy.ndarray
        3-by-3 array with the frame's axes. Each row is a vector.
    position : numpy.ndarray, optional
        3-element array with the frame's position. Default is [0, 0, 0].
    color : str or list of strings, optional
        Color of the frame. Default is None, which iterates over RGB.
    scale : float, optional
        Scale factor of the frame. Default is 1.0
    lw : float, optional
        Line width of the frame. Default is 1.5

    """
    if not hasattr(ax, 'plot'):
        raise TypeError("The given axis is not a 3D plot item.")
    colors = ([color]*3 if isinstance(color, str) else color) if color is not None else COLORS[:3]
    # Extract only the expected parameters from kwargs
    expected_params = {'scale': scale, 'lw': lw}
    for key in expected_params:
        if key in kwargs:
            expected_params[key] = kwargs[key]
    frame_coords = frame(dcm, position, scale)
    for axis in frame_coords:
        ax.plot(*axis, color=colors.pop(0), lw=lw)

def add_items(ax, **kwargs) -> None:
    """
    Add items to an existing 3D plot.

    Parameters
    ----------
    ax : mpl_toolkits.mplot3d.axes3d.Axes3D
        3D axis where the items will be added.
    kwargs : dict
        Dictionary with the items to be added. The keys are the items' types,
        and the values are the items' data and parameters.

    """
    if 'scatter' in kwargs:
        data = kwargs['scatter']
        if isinstance(data, (list, tuple, np.ndarray)):
            if isinstance(data, np.ndarray):
                data = data.T
            ax.scatter(*data)
        elif isinstance(data, dict):
            scatter_dict = copy.deepcopy(data)
            for k, v in scatter_dict.items():
                points = v.pop('data')    # Get N-by-3 numpy array from dict
                ax.scatter(*points.T, **v)
        else:
            raise TypeError(f"Unknown type for 'scatter': {type(data)}. Try an array, list or dict.")
    if 'lines' in kwargs:
        data = kwargs['lines']
        if isinstance(data, (list, tuple, np.ndarray)):
            if isinstance(data, np.ndarray):
                data = data.T
            ax.plot(*data)
        elif isinstance(data, dict):
            lines_dict = copy.deepcopy(data)
            for k, v in lines_dict.items():
                lines = v.pop('data')   # Get N-by-3 numpy array from dict
                ax.plot(*lines.T, **v)
        else:
            raise TypeError(f"Unknown type for 'lines': {type(data)}. Try an array, list or dict.")
    if 'frames' in kwargs:
        if isinstance(kwargs['frames'], dict):
            # Frames given as dictionary of dictionaries with each frame's attitude and position
            for k, v in kwargs['frames'].items():
                data = v.copy()
                add_frame(ax, dcm=data['attitude'], position=data.pop('position'), **data)
        elif isinstance(kwargs['frames'], list):
            # Frames given as list of frames.
            for frame_idx, frame_item in enumerate(kwargs['frames']):
                np_frame = np.copy(frame_item)
                if np_frame.shape == (3, 3):
                    add_frame(ax, dcm=np_frame)
                elif np_frame.shape == (3, 4):
                    add_frame(ax, dcm=np_frame[:, :3], position=np_frame[:, 3])
                else:
                    raise ValueError(f"Unknown shape for frame {frame_idx}: {np_frame.shape}")
        elif isinstance(kwargs['frames'], np.ndarray):
            # 3-dimensional array with the frames' data
            if kwargs['frames'].ndim == 2:
                np_frame = np.copy(kwargs['frames'])
                if np_frame.shape == (3, 3):
                    add_frame(ax, dcm=np_frame)
                elif np_frame.shape == (3, 4):
                    add_frame(ax, dcm=np_frame[:, :3], position=np_frame[:, 3])
                else:
                    raise ValueError(f"Unknown shape for frame: {np_frame.shape}")
            elif kwargs['frames'].ndim == 3:
                for frame_idx, frame_item in enumerate(kwargs['frames']):
                    np_frame = np.copy(frame_item)
                    if np_frame.shape == (3, 3):
                        add_frame(ax, dcm=np_frame)
                    elif np_frame.shape == (3, 4):
                        add_frame(ax, dcm=np_frame[:, :3], position=np_frame[:, 3])
                    else:
                        raise ValueError(f"Unknown shape for frame {frame_idx}: {np_frame.shape}")
            else:
                raise ValueError(f"Frames must be a 2D or 3D numpy array. Got shape {kwargs['frames'].shape}")
        else:
            raise TypeError(f"Unknown type for 'frames': {type(kwargs['frames'])}. Try a list or dict.")
    if 'ellipsoids' in kwargs:
        for k, v in kwargs['ellipsoids'].items():
            add_ellipsoid(ax, v, **v)

def plot3(**kwargs) -> None | tuple:
    """
    Plot 3-dimensional data in a cartesian coordinate system.

    Parameters
    ----------
    show : bool, optional
        Show the plot after creating it. Default is True. Otherwise, the plot
        Figure and Axes objects are returned.

    """
    # Build the plot
    fig = plt.figure()
    ax = fig.add_subplot(projection='3d')

    # Add items
    add_items(ax, **kwargs)

    # Set properties of plot
    plt.tight_layout()
    ax.set_xlabel('X')
    ax.set_ylabel('Y')
    ax.set_zlabel('Z')
    ax.set_aspect('equal')      # Added in matplotlib 3.6

    # Show or return the plot
    if not kwargs.get('show', True):
        return fig, ax
    plt.show()
